from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.options.legend.accessibility import LegendAccessibilityOptions
from highcharts_core.options.legend.navigation import LegendNavigation
from highcharts_core.options.legend.bubble_legend import BubbleLegend
from highcharts_core.options.legend.title import LegendTitle
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.shadows import ShadowOptions
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class Legend(HighchartsMeta):
    """The legend is a box containing a symbol and name for each series item or point
    item in the chart. Each series (or points in case of pie charts) is represented by
    a symbol and its name in the legend.

    .. seealso::

      It is possible to override the symbol creator function and create
      `custom legend symbols <https://jsfiddle.net/gh/get/library/pure/highcharts/highcharts/tree/master/samples/highcharts/studies/legend-custom-symbol/>`_.

    """

    def __init__(self, **kwargs):
        self._accessibility = None
        self._align = None
        self._align_columns = None
        self._background_color = None
        self._border_color = None
        self._border_width = None
        self._border_radius = None
        self._bubble_legend = None
        self._class_name = None
        self._enabled = None
        self._floating = None
        self._item_checkbox_style = None
        self._item_distance = None
        self._item_hidden_style = None
        self._item_hover_style = None
        self._item_margin_bottom = None
        self._item_margin_top = None
        self._item_style = None
        self._item_width = None
        self._label_format = None
        self._label_formatter = None
        self._layout = None
        self._margin = None
        self._max_height = None
        self._navigation = None
        self._padding = None
        self._reversed = None
        self._rtl = None
        self._shadow = None
        self._square_symbol = None
        self._symbol_height = None
        self._symbol_padding = None
        self._symbol_radius = None
        self._symbol_width = None
        self._title = None
        self._use_html = None
        self._vertical_align = None
        self._width = None
        self._x = None
        self._y = None

        self.accessibility = kwargs.get('accessibility', None)
        self.align = kwargs.get('align', None)
        self.align_columns = kwargs.get('align_columns', None)
        self.background_color = kwargs.get('background_color', None)
        self.border_color = kwargs.get('border_color', None)
        self.border_width = kwargs.get('border_width', None)
        self.border_radius = kwargs.get('border_radius', None)
        self.bubble_legend = kwargs.get('bubble_legend', None)
        self.class_name = kwargs.get('class_name', None)
        self.enabled = kwargs.get('enabled', None)
        self.floating = kwargs.get('floating', None)
        self.item_checkbox_style = kwargs.get('item_checkbox_style', None)
        self.item_distance = kwargs.get('item_distance', None)
        self.item_hidden_style = kwargs.get('item_hidden_style', None)
        self.item_hover_style = kwargs.get('item_hover_style', None)
        self.item_margin_bottom = kwargs.get('item_margin_bottom', None)
        self.item_margin_top = kwargs.get('item_margin_top', None)
        self.item_style = kwargs.get('item_style', None)
        self.item_width = kwargs.get('item_width', None)
        self.label_format = kwargs.get('label_format', None)
        self.label_formatter = kwargs.get('label_formatter', None)
        self.layout = kwargs.get('layout', None)
        self.margin = kwargs.get('margin', None)
        self.max_height = kwargs.get('max_height', None)
        self.navigation = kwargs.get('navigation', None)
        self.padding = kwargs.get('padding', None)
        self.reversed = kwargs.get('reversed', None)
        self.rtl = kwargs.get('rtl', None)
        self.shadow = kwargs.get('shadow', None)
        self.square_symbol = kwargs.get('square_symbol', None)
        self.symbol_height = kwargs.get('symbol_height', None)
        self.symbol_padding = kwargs.get('symbol_padding', None)
        self.symbol_radius = kwargs.get('symbol_radius', None)
        self.symbol_width = kwargs.get('symbol_width', None)
        self.title = kwargs.get('title', None)
        self.use_html = kwargs.get('use_html', None)
        self.vertical_align = kwargs.get('vertical_align', None)
        self.width = kwargs.get('width', None)
        self.x = kwargs.get('x', None)
        self.y = kwargs.get('y', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'legend'

    @property
    def accessibility(self) -> Optional[LegendAccessibilityOptions]:
        """Accessibility options for the legend.

        .. note::

          Requires the Accessibility module.

        :rtype: :class:`LegendAccessibilityOptions` or :obj:`None <python:None>`
        """
        return self._accessibility

    @accessibility.setter
    @class_sensitive(LegendAccessibilityOptions)
    def accessibility(self, value):
        self._accessibility = value

    @property
    def align(self) -> Optional[str]:
        """The horizontal alignment of the legend box within the chart area. Defaults to
        ``'center'``.

        Valid values are:

          * ``'left'``
          * ``'center'``
          * ``'right'``

        In the case that the legend is aligned in a corner position, the
        :meth:`Legend.layout` setting will determine whether to place it above/below or on
        the side of the plot area.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._align

    @align.setter
    def align(self, value):
        if not value:
            self._align = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['left', 'center', 'right']:
                raise errors.HighchartsValueError(f'align must be either "left", "center"'
                                                  f', or "right". Was: "{value}"')
            self._align = value

    @property
    def align_columns(self) -> Optional[bool]:
        """If the layout is horizontal and the legend items span over two lines or more,
        if the value is ``True`` will align the items into vertical columns.

        Setting this to ``False`` makes room for more items, but will look more messy.

        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._align_columns

    @align_columns.setter
    def align_columns(self, value):
        if value is None:
            self._align_columns = None
        else:
            self._align_columns = bool(value)

    @property
    def background_color(self) -> Optional[str | Gradient | Pattern]:
        """The background color or gradient for the legend. Defaults to
        :obj:`None <python:None>`.

        :returns: The backgorund color for the legend..
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        from highcharts_core import utility_functions
        self._background_color = utility_functions.validate_color(value)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The border color drawn around the legend. Defaults to
        ``'#999999'``.

        :returns: The color of the legend border.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_radius(self) -> Optional[int | float | Decimal]:
        """The border radius (in pixels) applied to the legend border. Defaults to
        ``0``.

        :returns: The border radius to apply to the legend border.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_radius

    @border_radius.setter
    def border_radius(self, value):
        self._border_radius = validators.numeric(value, allow_empty = True)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The border width (in pixels) applied to the legend border. Defaults to
        ``0``.

        :returns: The border width to apply to the legend border.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value, allow_empty = True)

    @property
    def bubble_legend(self) -> Optional[BubbleLegend]:
        """The bubble legend is an additional element in legend which presents the scale
        of the bubble series.

        Individual bubble ranges can be defined by user or calculated from series. In the
        case of automatically calculated ranges, a 1px margin of error is permitted.

        :rtype: :class:`BubbleLegend` or :obj:`None <python:None>`
        """
        return self._bubble_legend

    @bubble_legend.setter
    @class_sensitive(BubbleLegend)
    def bubble_legend(self, value):
        self._bubble_legend = value

    @property
    def class_name(self) -> Optional[str]:
        """A classname to apply styling using CSS.

        :returns: The classname to apply to enable styling via CSS.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, displays the legend. If ``False``, hides the legend. Defaults to
        ``False``.

        .. note::

          There is also a series-specific option,
          :meth:`PlotOptions.series.show_in_legend`, that can hide the series from the
          legend. In some series types this is ``False`` by default, so it must set to
          ``True`` in order to show the legend for the series.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def floating(self) -> Optional[bool]:
        """If ``True``, the plot area ignores the legend and can be rendered below the
        legend. If ``False``, the legend is rendered visually distinct (not overlapping)
        the plot area. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._floating

    @floating.setter
    def floating(self, value):
        if value is None:
            self._floating = None
        else:
            self._floating = bool(value)

    @property
    def item_checkbox_style(self) -> Optional[str]:
        """Default styling for the checkbox next to a legend item when
        :meth:`Legend.show_checkbox` is ``True``. Defaults to:
        ``'{"width": "13px", "height": "13px", "position":"absolute"}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._item_checkbox_style

    @item_checkbox_style.setter
    def item_checkbox_style(self, value):
        self._item_checkbox_style = validators.string(value, allow_empty = True)

    @property
    def item_distance(self) -> Optional[int | float | Decimal]:
        """In a legend with horizontal layout, the itemDistance defines the pixel distance
        between each item. Defaults to ``20``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._item_distance

    @item_distance.setter
    def item_distance(self, value):
        self._item_distance = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @property
    def item_hidden_style(self) -> Optional[str]:
        """Default styling for the legend item when the corresponding series or data
        point is hidden. Defaults to:
        ``'{"color": "#cccccc"}'``.

        .. warning::

          Only a subset of CSS is supported, notably those options related to text.

        .. note::

          Properties are inherited from :meth:`Legend.style` unless overridden here.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._item_hidden_style

    @item_hidden_style.setter
    def item_hidden_style(self, value):
        self._item_hidden_style = validators.string(value, allow_empty = True)

    @property
    def item_hover_style(self) -> Optional[str]:
        """Default styling for the legend item when the corresponding series or data
        point is in a hover state. Defaults to:
        ``'{"color": "#000000"}'``.

        .. warning::

          Only a subset of CSS is supported, notably those options related to text.

        .. note::

          Properties are inherited from :meth:`Legend.style` unless overridden here.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._item_hover_style

    @item_hover_style.setter
    def item_hover_style(self, value):
        self._item_hover_style = validators.string(value, allow_empty = True)

    @property
    def item_margin_bottom(self) -> Optional[int | float | Decimal]:
        """The bottom margin expressed in pixels for each legend item. Defaults to
        ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._item_margin_bottom

    @item_margin_bottom.setter
    def item_margin_bottom(self, value):
        self._item_margin_bottom = validators.numeric(value,
                                                      allow_empty = True,
                                                      minimum = 0)

    @property
    def item_margin_top(self) -> Optional[int | float | Decimal]:
        """The top margin expressed in pixels for each legend item. Defaults to
        ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._item_margin_top

    @item_margin_top.setter
    def item_margin_top(self, value):
        self._item_margin_top = validators.numeric(value,
                                                      allow_empty = True,
                                                      minimum = 0)

    @property
    def item_style(self) -> Optional[str]:
        """Default styling for each legend item. Defaults to:
        ``'{"color": "#333333", "cursor": "pointer", "fontSize": "12px", "fontWeight": "bold", "textOverflow": "ellipsis"}'``.

        .. warning::

          Only a subset of CSS is supported, notably those options related to text.
          However, a ``"width"`` property can be added to control the text width.

        .. note::

          The default ``"textOverflow"`` property makes long texts truncate. Set it to
          ``undefined`` to wrap text instead.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._item_style

    @item_style.setter
    def item_style(self, value):
        self._item_style = validators.string(value, allow_empty = True)

    @property
    def item_width(self) -> Optional[int | float | Decimal]:
        """The width for each legend item, expressed in pixels. Defaults to
        ``None``.

        By default, the items are laid out successively. In a horizontal layout, if the
        items are laid out across two rows or more, they will be vertically aligned
        depending on the :meth:`Legend.align_columns` setting.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._item_width

    @item_width.setter
    def item_width(self, value):
        self._item_width = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def label_format(self) -> Optional[str]:
        """A format string for each legend label. Defaults to:
        ``'{name}'``.

        .. note::

          Available variables relate to properties on the ``{{series}}``, or the
          ``{{point}}`` in the case of pie charts.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._label_format

    @label_format.setter
    def label_format(self, value):
        self._label_format = validators.string(value, allow_empty = True)

    @property
    def label_formatter(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function that formats each of the series' labels.

        .. note::

          The (JavaScript) ``this`` keyword refers to the series object for most chart
          types, and refers to the point object for pie charts.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._label_formatter

    @label_formatter.setter
    @class_sensitive(CallbackFunction)
    def label_formatter(self, value):
        self._label_formatter = value

    @property
    def layout(self) -> Optional[str]:
        """The layout of the legend items. Defaults to
        ``'horizontal'``.

        Accepts:

          * ``'horizontal'``
          * ``'vertical'``
          * ``'proximate'``

        When proximate, the legend items will be placed as close as possible to the graphs
        they're representing, except in inverted charts or when the legend position does
        not allow it.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._layout

    @layout.setter
    def layout(self, value):
        if not value:
            self._layout = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['horizontal', 'vertical', 'proximate']:
                raise errors.HighchartsValueError(f'layout must be either "horizontal", '
                                                  f', "vertical", or "proximate". Was: '
                                                  f'"{value}"')
            self._layout = value

    @property
    def margin(self) -> Optional[int | float | Decimal]:
        """If the plot area sized is calculated automatically and the legend is not
        floating, the legend margin is the space between the legend and the axis labels or
        plot area. Defaults to ``12``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._margin

    @margin.setter
    def margin(self, value):
        self._margin = validators.numeric(value,
                                          allow_empty = True)

    @property
    def max_height(self) -> Optional[int | float | Decimal]:
        """The maximum height for the legend, expressed in pixels. Defaults to
        ``None``.

        When the maximum height is extended, navigation will show.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max_height

    @max_height.setter
    def max_height(self, value):
        self._max_height = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def navigation(self) -> Optional[LegendNavigation]:
        """Options for paging or navigation within the legend when the legend overflows
        the available space.

        .. hint::

          Navigation works well on screen, but not in static exported images. One way of
          working around that is to increase the chart height in export.

        :rtype: :class:`LegendNavigation` or :obj:`None <python:None>`
        """
        return self._navigation

    @navigation.setter
    @class_sensitive(LegendNavigation)
    def navigation(self, value):
        self._navigation = value

    @property
    def padding(self) -> Optional[int | float | Decimal]:
        """The inner padding of the legend box. Defaults to
        ``8``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._padding

    @padding.setter
    def padding(self, value):
        self._padding = validators.numeric(value, allow_empty = True)

    @property
    def reversed(self) -> Optional[bool]:
        """If ``True``, reverses the order of the legend items compared to the order of
        series/points as defined in the configuration object.
        Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._reversed

    @reversed.setter
    def reversed(self, value):
        if value is None:
            self._reversed = None
        else:
            self._reversed = bool(value)

    @property
    def rtl(self) -> Optional[bool]:
        """If ``True``, displays the symbol on the right side of the text rather than
        the left (this is common in RTL languages like Arabic or Hebrew). Defaults to
        ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._rtl

    @rtl.setter
    def rtl(self, value):
        if value is None:
            self._rtl = None
        else:
            self._rtl = bool(value)

    @property
    def shadow(self) -> Optional[bool | ShadowOptions]:
        """Configuration of a drop shadow applied to the legend box. Accepts either
        a boolean value of ``False`` which disables any shadow, or a
        :class:`ShadowOptions` instance with the applicable configuration.

        Defaults to ``False``.

        .. warning::

          Requires that :meth:`Legend.background_color` be set.

        :rtype: :class:`bool <python:bool>` or :class:`ShadowOptions` or
          :obj:`None <python:None>`
        """
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        if value is None:
            self._shadow = None
        elif value is False:
            self._shadow = False
        else:
            if value is True:
                value = ShadowOptions(enabled = True)
            else:
                value = validate_types(value,
                                       types = ShadowOptions,
                                       allow_none = False)
            self._shadow = value

    @property
    def square_symbol(self) -> Optional[bool]:
        """If ``True``, the legend symbol width will be the same as the symbol height,
        which in turn defaults to the font size of the legend items. Defaults to
        ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._square_symbol

    @square_symbol.setter
    def square_symbol(self, value):
        if value is None:
            self._square_symbol = None
        else:
            self._square_symbol = bool(value)

    @property
    def symbol_height(self) -> Optional[int | float | Decimal]:
        """The height of the symbol (in pixels) for series types that use a rectangle in
        the legend. Defaults to the font size of legend items.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._symbol_height

    @symbol_height.setter
    def symbol_height(self, value):
        self._symbol_height = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @property
    def symbol_padding(self) -> Optional[int | float | Decimal]:
        """The pixel padding between the legend item symbol and the item text. Defaults
        to ``5``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._symbol_padding

    @symbol_padding.setter
    def symbol_padding(self, value):
        self._symbol_padding = validators.numeric(value, allow_empty = True)

    @property
    def symbol_radius(self) -> Optional[int | float | Decimal]:
        """The border radius of symbol for series types that use a rectangle in the
        legend. Defaults to one half of the :class:`Legend.symbol_height`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._symbol_radius

    @symbol_radius.setter
    def symbol_radius(self, value):
        self._symbol_radius = validators.numeric(value, allow_empty = True)

    @property
    def symbol_width(self) -> Optional[int | float | Decimal]:
        """The width of the legend item symbol (in pixels). When
        :meth:`Legend.square_symbol` is ``True``, this defaults to the
        :meth:`Legend.symbol_height`, otherwise ``16``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._symbol_width

    @symbol_width.setter
    def symbol_width(self, value):
        self._symbol_width = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def title(self) -> Optional[LegendTitle]:
        """A title to be added on top of the legend.

        :rtype: :class:`LegendTitle` or :obj:`None <python:None>`
        """
        return self._title

    @title.setter
    @class_sensitive(LegendTitle)
    def title(self, value):
        self._title = value

    @property
    def use_html(self) -> Optional[bool]:
        """If ``True``, will use HTML to render the legend item texts. If ``False``, will
        use SVG or WebGL as applicable.

        Defaults to ``False``.

        :returns: Flag indicating whether to render annotation labels using HTML.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._use_html

    @use_html.setter
    def use_html(self, value):
        if value is None:
            self._use_html = None
        else:
            self._use_html = bool(value)

    @property
    def vertical_align(self) -> Optional[str]:
        """The vertical alignment of the legend box. Defaults to
        ``'bottom'``.

        Accepts:

          * ``'bottom'``
          * ``'middle'``
          * ``'top'``

        .. hint::

          Vertical position can be further determined by the :meth:`Legend.y` option.

        .. note::

          If the legend is aligned in a corner position, the :meth:`Legend.layout` setting
          will determine whether to place it above/below or on the side of the plot area.

          When the layout option is ``'proximate'``, the ``vertical_align`` option
          does not apply.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._vertical_align

    @vertical_align.setter
    def vertical_align(self, value):
        if not value:
            self._vertical_align = None
        else:
            value = validators.string(value, allow_empty = True)
            value = value.lower()
            if value not in ['bottom', 'middle', 'top']:
                raise errors.HighchartsValueError(f'vertical_align expects either "top", '
                                                  f'"middle", or "bottom". Was: {value}')
            self._vertical_align = value

    @property
    def width(self) -> Optional[str | int | float | Decimal]:
        """The width of the legend box. If expressed as a numeric value, the value is in
        pixels. If as a :class:`str <python:str>`, then accepts a value expressed as a
        percentage of the chart area (e.g. ``'40%'``).

        If :obj:`None <python:None>`, then defaults to the full chart width if above or
        below the chart and half the chart width aligned to the left or right of the
        chart.

        Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        if not value:
            self._width = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError(f'if width is a string, it is '
                                                      f'expected to be a percentage of '
                                                      f'the chart area. No % sign found '
                                                      f'in value: {value}')
                self._width = value
            except (TypeError, ValueError):
                self._width = validators.numeric(value, minimum = 0)

    @property
    def x(self) -> Optional[int]:
        """The x position offset of the legend relative to its horizontal alignment
        (:meth:`Legend.align`) within :meth:`Chart.spacing_left` and
        :meth:`Chart.spacing_right`. Defaults to ``0``.

        .. note::

          Negative values move it to the left, positive values move it to the right.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._x

    @x.setter
    def x(self, value):
        self._x = validators.numeric(value, allow_empty = True)

    @property
    def y(self) -> Optional[int]:
        """The y position offset of the legend relative to its vertical alignment
        (:meth:`Legend.vertical_align`) within :meth:`Chart.spacing_top` and
        :meth:`Chart.spacing_bottom`. Defaults to ``0``.

        .. note::

          Negative values move it up, positive values move it down.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._y

    @y.setter
    def y(self, value):
        self._y = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'align': as_dict.get('align', None),
            'align_columns': as_dict.get('alignColumns', None),
            'background_color': as_dict.get('backgroundColor', None),
            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'border_radius': as_dict.get('borderRadius', None),
            'bubble_legend': as_dict.get('bubbleLegend', None),
            'class_name': as_dict.get('className', None),
            'enabled': as_dict.get('enabled', None),
            'floating': as_dict.get('floating', None),
            'item_checkbox_style': as_dict.get('itemCheckboxStyle', None),
            'item_distance': as_dict.get('itemDistance', None),
            'item_hidden_style': as_dict.get('itemHiddenStyle', None),
            'item_hover_style': as_dict.get('itemHoverStyle', None),
            'item_margin_bottom': as_dict.get('itemMarginBottom', None),
            'item_margin_top': as_dict.get('itemMarginTop', None),
            'item_style': as_dict.get('itemStyle', None),
            'item_width': as_dict.get('itemWidth', None),
            'label_format': as_dict.get('labelFormat', None),
            'label_formatter': as_dict.get('labelFormatter', None),
            'layout': as_dict.get('layout', None),
            'margin': as_dict.get('margin', None),
            'max_height': as_dict.get('maxHeight', None),
            'navigation': as_dict.get('navigation', None),
            'padding': as_dict.get('padding', None),
            'reversed': as_dict.get('reversed', None),
            'rtl': as_dict.get('rtl', None),
            'shadow': as_dict.get('shadow', None),
            'square_symbol': as_dict.get('squareSymbol', None),
            'symbol_height': as_dict.get('symbolHeight', None),
            'symbol_padding': as_dict.get('symbolPadding', None),
            'symbol_radius': as_dict.get('symbolRadius', None),
            'symbol_width': as_dict.get('symbolWidth', None),
            'title': as_dict.get('title', None),
            'use_html': as_dict.get('useHTML', None),
            'vertical_align': as_dict.get('verticalAlign', None),
            'width': as_dict.get('width', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'accessibility': self.accessibility,
            'align': self.align,
            'alignColumns': self.align_columns,
            'backgroundColor': self.background_color,
            'borderColor': self.border_color,
            'borderWidth': self.border_width,
            'borderRadius': self.border_radius,
            'bubbleLegend': self.bubble_legend,
            'className': self.class_name,
            'enabled': self.enabled,
            'floating': self.floating,
            'itemCheckboxStyle': self.item_checkbox_style,
            'itemDistance': self.item_distance,
            'itemHiddenStyle': self.item_hidden_style,
            'itemHoverStyle': self.item_hover_style,
            'itemMarginBottom': self.item_margin_bottom,
            'itemMarginTop': self.item_margin_top,
            'itemStyle': self.item_style,
            'itemWidth': self.item_width,
            'labelFormat': self.label_format,
            'labelFormatter': self.label_formatter,
            'layout': self.layout,
            'margin': self.margin,
            'maxHeight': self.max_height,
            'navigation': self.navigation,
            'padding': self.padding,
            'reversed': self.reversed,
            'rtl': self.rtl,
            'shadow': self.shadow,
            'squareSymbol': self.square_symbol,
            'symbolHeight': self.symbol_height,
            'symbolPadding': self.symbol_padding,
            'symbolRadius': self.symbol_radius,
            'symbolWidth': self.symbol_width,
            'title': self.title,
            'useHTML': self.use_html,
            'verticalAlign': self.vertical_align,
            'width': self.width,
            'x': self.x,
            'y': self.y
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.credits import Credits as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'enabled': True,
      'href': 'https://www.somewhere.com',
      'text': 'Test Text for the Credits label'
    }, None),
    ({
      'enabled': True,
      'href': 'https://www.somewhere.com',
      'style': {
          'color': '#cccccc',
          'fontSize': '12px'
      },
      'text': 'Test Text for the Credits label'
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('credits/01.js', False, None),
    ('credits/02.js', False, None),

    ('credits/error-01.js', False, (errors.HighchartsValueError,
                                    errors.HighchartsParseError,
                                    JSONDecodeError,
                                    ValueError)),
    ('credits/error-02.js', False, (errors.HighchartsValueError,
                                    errors.HighchartsParseError,
                                    JSONDecodeError,
                                    ValueError)),

    ('credits/01.js', True, None),
    ('credits/02.js', True, None),

    ('credits/error-01.js', True, (errors.HighchartsValueError,
                                   errors.HighchartsParseError,
                                   JSONDecodeError,
                                   ValueError)),
    ('credits/error-02.js', True, (errors.HighchartsValueError,
                                   errors.HighchartsParseError,
                                   JSONDecodeError,
                                   ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta


class LinearGradient(HighchartsMeta):
    """Configuration of a linear gradient pattern."""

    def __init__(self, **kwargs):
        self._x1 = None
        self._x2 = None
        self._y1 = None
        self._y2 = None

        self.x1 = kwargs.get('x1', None)
        self.x2 = kwargs.get('x2', None)
        self.y1 = kwargs.get('y1', None)
        self.y2 = kwargs.get('y2', None)

    @property
    def x1(self) -> Optional[float]:
        """Start horizontal position for the gradient. Expects a value between 0 - 1.

        :rtype: :class:`float <python:float>` or :obj:`None <python:None>`
        """
        return self._x1

    @x1.setter
    def x1(self, value):
        self._x1 = validators.float(value,
                                    allow_empty = True,
                                    minimum = 0,
                                    maximum = 1)

    @property
    def x2(self) -> Optional[float]:
        """End horizontal position for the gradient. Expects a value between 0 - 1.

        :rtype: :class:`float <python:float>` or :obj:`None <python:None>`
        """
        return self._x2

    @x2.setter
    def x2(self, value):
        self._x2 = validators.float(value,
                                    allow_empty = True,
                                    minimum = 0,
                                    maximum = 1)

    @property
    def y1(self) -> Optional[float]:
        """Start vertical position for the gradient. Expects a value between 0 - 1.

        :rtype: :class:`float <python:float>` or :obj:`None <python:None>`
        """
        return self._y1

    @y1.setter
    def y1(self, value):
        self._y1 = validators.float(value,
                                    allow_empty = True,
                                    minimum = 0,
                                    maximum = 1)

    @property
    def y2(self) -> Optional[float]:
        """End vertical position for the gradient. Expects a value between 0 - 1.

        :rtype: :class:`float <python:float>` or :obj:`None <python:None>`
        """
        return self._y2

    @y2.setter
    def y2(self, value):
        self._y2 = validators.float(value,
                                    allow_empty = True,
                                    minimum = 0,
                                    maximum = 1)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'x1': as_dict.get('x1', None),
            'x2': as_dict.get('x2', None),
            'y1': as_dict.get('y1', None),
            'y2': as_dict.get('y2', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'x1': self.x1,
            'x2': self.x2,
            'y1': self.y1,
            'y2': self.y2
        }

        return untrimmed


class RadialGradient(HighchartsMeta):
    """Configuration of a radial gradient pattern."""

    def __init__(self, **kwargs):
        self._cx = None
        self._cy = None
        self._r = None

        self.cx = kwargs.get('cx', None)
        self.cy = kwargs.get('cy', None)
        self.r = kwargs.get('r', None)

    @property
    def cx(self) -> Optional[float]:
        """Center position for the gradient along the horizontal axis. Expects a value
        between 0 - 1, relative to the shape.

        :rtype: :class:`float <python:float>` or :obj:`None <python:None>`
        """
        return self._cx

    @cx.setter
    def cx(self, value):
        self._cx = validators.float(value,
                                    allow_empty = True,
                                    minimum = 0,
                                    maximum = 1)

    @property
    def cy(self) -> Optional[float]:
        """Center position for the gradient along the vertical axis. Expects a value
        between 0 - 1, relative to the shape.

        :rtype: :class:`float <python:float>` or :obj:`None <python:None>`
        """
        return self._cy

    @cy.setter
    def cy(self, value):
        self._cy = validators.float(value,
                                    allow_empty = True,
                                    minimum = 0,
                                    maximum = 1)

    @property
    def r(self) -> Optional[float]:
        """Radius of the gradient relative to the shape. Expects a value between 0 - 1.

        :rtype: :class:`float <python:float>` or :obj:`None <python:None>`
        """
        return self._r

    @r.setter
    def r(self, value):
        self._r = validators.float(value,
                                   allow_empty = True,
                                   minimum = 0,
                                   maximum = 1)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'cx': as_dict.get('cx', None),
            'cy': as_dict.get('cy', None),
            'r': as_dict.get('r', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'cx': self.cx,
            'cy': self.cy,
            'r': self.r
        }

        return untrimmed


class Gradient(HighchartsMeta):
    """Configuration of a gradient color pattern."""

    def __init__(self, **kwargs):
        self._linear_gradient = None
        self._radial_gradient = None
        self._stops = None

        self.linear_gradient = kwargs.get('linear_gradient', None)
        self.radial_gradient = kwargs.get('radial_gradient', None)
        self.stops = kwargs.get('stops', None)

    @property
    def linear_gradient(self) -> Optional[LinearGradient]:
        """Holds an object that defines the start position and end position of the
        gradient relative to the shape.

        :rtype: :class:`LinearGradient` or :obj:`None <python:None>`
        """
        return self._linear_gradient

    @linear_gradient.setter
    @class_sensitive(LinearGradient)
    def linear_gradient(self, value):
        self._linear_gradient = value

    @property
    def radial_gradient(self) -> Optional[RadialGradient]:
        """Holds an object that defines the center and radius of a radial gradient
        relative to the shape.

        :rtype: :class:`RadialGradient` or :obj:`None <python:None>`
        """
        return self._radial_gradient

    @radial_gradient.setter
    @class_sensitive(RadialGradient)
    def radial_gradient(self, value):
        self._radial_gradient = value

    @property
    def stops(self):
        """Definition of the stop points within the color gradient.

        The value expected is an iterable of two-value iterables, where the first value
        is the start position of the gradient, the second is the color to apply to this
        section of the gradient.

        :rtype: :class:`list <python:list>` of 2-item :class:`list <python:list>` where
          item 1 is a numeric, and item 2 is a :class:`str <python:str>`

        """
        return self._stops

    @stops.setter
    def stops(self, value):
        value = validators.iterable(value, allow_empty = True)
        if not value:
            self._stops = None
        else:
            stops = []
            for item in value:
                stop = validators.iterable(item,
                                           allow_empty = False,
                                           minimum_length = 2,
                                           maximum_length = 2)
                stop[0] = validators.numeric(stop[0],
                                             allow_empty = False)
                stop[1] = validators.string(stop[1],
                                            allow_empty = False)
                stops.append(stop)

            self._stops = stops

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'linear_gradient': as_dict.get('linearGradient', None),
            'radial_gradient': as_dict.get('radialGradient', None),
            'stops': as_dict.get('stops', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'linearGradient': self.linear_gradient,
            'radialGradient': self.radial_gradient,
            'stops': self.stops
        }

        return untrimmed



"""Unit tests for ``highcharts/utility_functions``"""

import pytest

from abc import ABC, abstractmethod

from highcharts_core import utility_functions


@pytest.mark.parametrize('kwargs, expected_column_names, expected_records, error', [
    ({
        'csv_data': "Date,Header\r\n01/01/2023,2\r\n01/02/2023,4\r\n01/03/2023,8"
     }, ['Date', 'HeadCount'], 3, None),
])
def test_parse_csv(kwargs, expected_column_names, expected_records, error):
    if not error:
        columns, records_as_dicts = utility_functions.parse_csv(**kwargs)
        assert columns is not None
        assert len(columns) == len(expected_column_names)
        
        assert records_as_dicts is not None
        assert len(records_as_dicts) == expected_records
    else:
        with pytest.raises(error):
            result = utility_functions.parse_csv(**kwargs)


from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core.metaclasses import HighchartsMeta


class AxisBreak(HighchartsMeta):
    """The definition of a break (non-continuous section, typically indicating a "jump")
    in the axis.

    .. note::

      The sections defined as breaks will be left out and all the points shifted
      closer to each other.

    """

    def __init__(self, **kwargs):
        self._break_size = None
        self._from_ = None
        self._repeat = None
        self._to = None

        self.break_size = kwargs.get('break_size', None)
        self.from_ = kwargs.get('from_', None)
        self.repeat = kwargs.get('repeat', None)
        self.to = kwargs.get('to', None)

    @property
    def break_size(self) -> Optional[int | float | Decimal]:
        """A number indicating how much space should be left between the start and the end
        of the break. Defaults to ``0``.

        .. note::

          The break size is given in axis units, so for instance on a datetime axis, a
          break size of ``3600000`` would indicate the equivalent of an hour.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._break_size

    @break_size.setter
    def break_size(self, value):
        self._break_size = validators.numeric(value, allow_empty = True)

    @property
    def from_(self) -> Optional[int | float | Decimal]:
        """The point where the break starts. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._from_

    @from_.setter
    def from_(self, value):
        self._from_ = validators.numeric(value, allow_empty = True)

    @property
    def repeat(self) -> Optional[int | float | Decimal]:
        """Defines an interval after which the break appears again. Defaults to ``0``.

        .. note::

          If ``0`` or :obj:`None <python:None>`, the breaks do not repeat.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._repeat

    @repeat.setter
    def repeat(self, value):
        self._repeat = validators.numeric(value,
                                          allow_empty = True,
                                          minimum = 0)

    @property
    def to(self) -> Optional[int | float | Decimal]:
        """The point where the break ends. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._to

    @to.setter
    def to(self, value):
        self._to = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'break_size': as_dict.get('breakSize', None),
            'from_': as_dict.get('from', None),
            'repeat': as_dict.get('repeat', None),
            'to': as_dict.get('to', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'breakSize': self.break_size,
            'from': self.from_,
            'repeat': self.repeat,
            'to': self.to
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class LegendTitle(HighchartsMeta):
    """A title to be added on top of the legend."""

    def __init__(self, **kwargs):
        self._style = None
        self._text = None

        self.style = kwargs.get('style', None)
        self.text = kwargs.get('text', None)

    @property
    def style(self) -> Optional[str]:
        """CSS styling to apply to the title. Defaults to
        ``'{constants.DEFAULT_LEGEND.get('title', {}).get('style')}'``.

        :rtype: :class:`str` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.string(value, allow_empty = True, coerce_value = True)

    @property
    def text(self) -> Optional[str]:
        """A text or HTML string for the tite.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._text

    @text.setter
    def text(self, value):
        self._text = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'style': as_dict.get('style', None),
            'text': as_dict.get('text', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'style': self.style,
            'text': self.text
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core.metaclasses import HighchartsMeta


class DataSorting(HighchartsMeta):
    """Options for the series data sorting."""

    def __init__(self, **kwargs):
        self._enabled = None
        self._match_by_name = None
        self._sort_key = None

        self.enabled = kwargs.get('enabled', None)
        self.match_by_name = kwargs.get('match_by_name', None)
        self.sort_key = kwargs.get('sort_key', None)

    @property
    def enabled(self) -> Optional[bool]:
        """Enable or disable data sorting for the series. Defaults to
        :obj:`None <python:None>` (equivalent to ``False``).

        .. hint::

          Use :meth:`XAxis.reversed` to change the sorting order.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def match_by_name(self) -> Optional[bool]:
        """If ``True``, allow matching points by name in an update. If ``False``, points
        will be matched by order. Defaults to :obj:`None <python:None>` (equivalent to
        ``False``).

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._match_by_name

    @match_by_name.setter
    def match_by_name(self, value):
        if value is None:
            self._match_by_name = None
        else:
            self._match_by_name = bool(value)

    @property
    def sort_key(self) -> Optional[str]:
        """Determines what data value should be used to sort by. Defaults to
        :obj:`None <python:None>`, which behaves as if the value were ``'y'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._sort_key

    @sort_key.setter
    def sort_key(self, value):
        self._sort_key = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'enabled': as_dict.get('enabled', None),
            'match_by_name': as_dict.get('matchByName', None),
            'sort_key': as_dict.get('sortKey', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'enabled': self.enabled,
            'matchByName': self.match_by_name,
            'sortKey': self.sort_key
        }

        return untrimmed



import pytest

from validator_collection import checkers

from highcharts_core import utility_functions
from highcharts_core.metaclasses import HighchartsMeta


class Grandparent(HighchartsMeta):
    def __init__(self, **kwargs):
        self._item1 = None
        self._item2 = None

        self.item1 = kwargs.get('item1', 123)
        self.item2 = kwargs.get('item2', 345)

    @property
    def item1(self):
        return self._item1

    @item1.setter
    def item1(self, value):
        self._item1 = value

    @property
    def item2(self):
        return self._item2

    @item2.setter
    def item2(self, value):
        self._item2 = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        return as_dict

    def _to_untrimmed_dict(self, in_cls = None):
        untrimmed = {
            'item1': self.item1,
            'item2': self.item2
        }

        return untrimmed

    def return_value(self):
        return 'Grandparent'


class ParentA(Grandparent):
    def __init__(self, **kwargs):
        self._item3 = None
        self._item4 = None

        self.item3 = kwargs.get('item3', 567)
        self.item4 = kwargs.get('item4', 890)

        super().__init__(**kwargs)

    @property
    def item3(self):
        return self._item3

    @item3.setter
    def item3(self, value):
        self._item3 = value

    @property
    def item4(self):
        return self._item4

    @item4.setter
    def item4(self, value):
        self._item4 = value

    def _to_untrimmed_dict(self, in_cls = None):
        print(f'starting {self.__class__.__name__}._to_untrimmed_dict()')
        untrimmed = {
            'item3': self.item3,
            'item4': self.item4
        }
        parent_as_dict = utility_functions.mro__to_untrimmed_dict(self,
                                                                  in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        print(f'ending {self.__class__.__name__}._to_untrimmed_dict()')

        return untrimmed

    def return_value(self):
        return 'ParentA'


class ParentB(Grandparent):
    def __init__(self, **kwargs):
        self._item5 = None
        self._item6 = None

        self.item5 = kwargs.get('item5', 987)
        self.item6 = kwargs.get('item6', 654)

        super().__init__(**kwargs)

    @property
    def item5(self):
        return self._item5

    @item5.setter
    def item5(self, value):
        self._item5 = value

    @property
    def item6(self):
        return self._item6

    @item6.setter
    def item6(self, value):
        self._item6 = value

    def _to_untrimmed_dict(self, in_cls = None):
        print(f'starting {self.__class__.__name__}._to_untrimmed_dict()')
        untrimmed = {
            'item5': self.item5,
            'item6': self.item6
        }
        parent_as_dict = utility_functions.mro__to_untrimmed_dict(self,
                                                                  in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        print(f'ending  {self.__class__.__name__}._to_untrimmed_dict()')

        return untrimmed

    def return_value(self):
        return 'ParentB'


class Child(ParentA, ParentB):
    def __init__(self, **kwargs):
        self._item7 = None

        self.item7 = kwargs.get('item7', 321)

        super().__init__(**kwargs)

    @property
    def item7(self):
        return self._item7

    @item7.setter
    def item7(self, value):
        self._item7 = value

    def _to_untrimmed_dict(self, in_cls = None):
        print(f'starting {self.__class__.__name__}._to_untrimmed_dict()')
        untrimmed = {
            'item7': self.item7
        }
        parent_as_dict = utility_functions.mro__to_untrimmed_dict(self,
                                                                  in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        print(f'ending {self.__class__.__name__}._to_untrimmed_dict()')

        return untrimmed

    def return_value(self):
        return 'Child'


class GrandChild(Child):
    def __init__(self, **kwargs):
        self._item8 = None

        self.item8 = kwargs.get('item8', 12)

        super().__init__(**kwargs)

    @property
    def item8(self):
        return self._item8

    @item8.setter
    def item8(self, value):
        self._item8 = value

    def _to_untrimmed_dict(self, in_cls = None):
        untrimmed = {
            'item8': self.item8
        }

        parent_as_dict = utility_functions.mro__to_untrimmed_dict(self,
                                                                  in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        print('consolidated parent_as_dict:')
        print(parent_as_dict)

        print(f'ending {self.__class__.__name__}._to_untrimmed_dict()')

        return untrimmed

    def return_value(self):
        return 'GrandChild'


def mro_return(obj, exclude_self = False):
    cls = obj.__class__
    mro = cls.mro()
    if exclude_self:
        result = [x.return_value(obj) for x in mro
                  if hasattr(x, 'return_value') and
                  x != cls]
    else:
        result = [x.return_value(obj) for x in mro
                  if hasattr(x, 'return_value')]

    return result


@pytest.mark.parametrize('cls, kwargs, error', [
    (GrandChild, {}, None),
    (Child, {}, None),
])
def test__init(cls, kwargs, error):
    if not error:
        result = cls(**kwargs)
        assert isinstance(result, cls) is True
        assert isinstance(result, Grandparent) is True
        assert isinstance(result, HighchartsMeta) is True

        if isinstance(result, GrandChild):
            assert isinstance(result, Child) is True
            assert isinstance(result, ParentA) is True
            assert isinstance(result, ParentB) is True
            assert isinstance(result, Grandparent) is True
            assert hasattr(result, 'item8') is True
            assert result.item8 == kwargs.get('item8', 12)
            assert isinstance(result, ParentA) is True
            assert isinstance(result, ParentB) is True
            assert isinstance(result, Grandparent) is True
            assert hasattr(result, 'item7') is True
            assert result.item7 == kwargs.get('item7', 321)
            assert isinstance(result, Grandparent) is True
            assert hasattr(result, 'item3') is True
            assert hasattr(result, 'item4') is True
            assert result.item3 == kwargs.get('item3', 567)
            assert result.item4 == kwargs.get('item4', 890)
        if isinstance(result, Child):
            assert isinstance(result, ParentA) is True
            assert isinstance(result, ParentB) is True
            assert isinstance(result, Grandparent) is True
            assert hasattr(result, 'item7') is True
            assert result.item7 == kwargs.get('item7', 321)
        if isinstance(result, ParentA):
            assert isinstance(result, Grandparent) is True
            assert hasattr(result, 'item3') is True
            assert hasattr(result, 'item4') is True
            assert result.item3 == kwargs.get('item3', 567)
            assert result.item4 == kwargs.get('item4', 890)
        if isinstance(result, ParentB):
            assert isinstance(result, Grandparent) is True
            assert hasattr(result, 'item5') is True
            assert hasattr(result, 'item6') is True
            assert result.item5 == kwargs.get('item5', 987)
            assert result.item6 == kwargs.get('item6', 654)
        if isinstance(result, Grandparent):
            assert isinstance(result, HighchartsMeta) is True
            assert hasattr(result, 'item1') is True
            assert hasattr(result, 'item2') is True
            assert result.item1 == kwargs.get('item1', 123)
            assert result.item2 == kwargs.get('item2', 345)

    else:
        with pytest.raises(error):
            result = cls(**kwargs)


@pytest.mark.parametrize('cls, kwargs, expected, error', [
    (GrandChild,
     {},
     {
         'item1': 123,
         'item2': 345,
         'item3': 567,
         'item4': 890,
         'item5': 987,
         'item6': 654,
         'item7': 321,
         'item8': 12
     },
     None),
    (Child,
     {},
     {
         'item1': 123,
         'item2': 345,
         'item3': 567,
         'item4': 890,
         'item5': 987,
         'item6': 654,
         'item7': 321
     },
     None),
    (ParentB,
     {},
     {
         'item1': 123,
         'item2': 345,
         'item5': 987,
         'item6': 654,
     },
     None),
    (Grandparent,
     {},
     {
         'item1': 123,
         'item2': 345,
     },
     None),
])
def test__to_untrimmed_dict(cls, kwargs, expected, error):
    if not error:
        instance = cls(**kwargs)
        result = instance._to_untrimmed_dict()

        if isinstance(instance, GrandChild) is True:
            assert len(result) == 8

            assert 'item8' in result
            assert 'item7' in result
            assert 'item6' in result
            assert 'item5' in result
            assert 'item4' in result
            assert 'item3' in result
            assert 'item2' in result
            assert 'item1' in result

            assert result.get('item8') == kwargs.get('item8', 12) == instance.item8
            assert result.get('item7') == kwargs.get('item7', 321) == instance.item7
            assert result.get('item6') == kwargs.get('item6', 654) == instance.item6
            assert result.get('item5') == kwargs.get('item5', 987) == instance.item5
            assert result.get('item4') == kwargs.get('item4', 890) == instance.item4
            assert result.get('item3') == kwargs.get('item3', 567) == instance.item3
            assert result.get('item2') == kwargs.get('item2', 345) == instance.item2
            assert result.get('item1') == kwargs.get('item1', 123) == instance.item1

        if isinstance(instance, Child) is True and not isinstance(instance, GrandChild):
            print(result)
            assert len(result) == 7

            assert 'item8' not in result
            assert 'item7' in result
            assert 'item6' in result
            assert 'item5' in result
            assert 'item4' in result
            assert 'item3' in result
            assert 'item2' in result
            assert 'item1' in result

            assert result.get('item7') == kwargs.get('item7', 321) == instance.item7
            assert result.get('item6') == kwargs.get('item6', 654) == instance.item6
            assert result.get('item5') == kwargs.get('item5', 987) == instance.item5
            assert result.get('item4') == kwargs.get('item4', 890) == instance.item4
            assert result.get('item3') == kwargs.get('item3', 567) == instance.item3
            assert result.get('item2') == kwargs.get('item2', 345) == instance.item2
            assert result.get('item1') == kwargs.get('item1', 123) == instance.item1

        if isinstance(instance, ParentB) is True and not isinstance(instance, Child):
            print(result)
            assert len(result) == 4

            assert 'item8' not in result
            assert 'item7' not in result
            assert 'item6' in result
            assert 'item5' in result
            assert 'item4' not in result
            assert 'item3' not in result
            assert 'item2' in result
            assert 'item1' in result

            assert result.get('item6') == kwargs.get('item6', 654) == instance.item6
            assert result.get('item5') == kwargs.get('item5', 987) == instance.item5
            assert result.get('item2') == kwargs.get('item2', 345) == instance.item2
            assert result.get('item1') == kwargs.get('item1', 123) == instance.item1

        assert checkers.are_dicts_equivalent(result, expected) is True
    else:
        with pytest.raises(error):
            instance = cls(**kwargs)
            result = instance._to_untrimmed_dict()


@pytest.mark.parametrize('cls, exclude_self', [
    (Child, False),
    (ParentA, False),
    (ParentB, False),
    (GrandChild, False),
    (Grandparent, False),

    (Child, True),
    (ParentA, True),
    (ParentB, True),
    (GrandChild, True),
    (Grandparent, True),
])
def test_mro_return(cls, exclude_self):
    instance = cls()
    result = mro_return(instance, exclude_self = exclude_self)
    if instance.__class__.__name__ == 'Child':
        if exclude_self:
            assert len(result) == 3
            assert 'Child' not in result
        else:
            assert len(result) == 4
            assert 'Child' in result
        assert 'ParentA' in result
        assert 'ParentB' in result
        assert 'Grandparent' in result
    elif instance.__class__.__name__ == 'ParentA':
        if exclude_self:
            assert len(result) == 1
            assert 'ParentA' not in result
        else:
            assert len(result) == 2
            assert 'ParentA' in result
        assert 'Grandparent' in result
    elif instance.__class__.__name__ == 'ParentB':
        if exclude_self:
            assert len(result) == 1
            assert 'ParentB' not in result
        else:
            assert len(result) == 2
            assert 'ParentB' in result
        assert 'Grandparent' in result
    elif instance.__class__.__name__ == 'Grandparent':
        if exclude_self:
            assert len(result) == 0
            assert 'Grandparent' not in result
        else:
            assert len(result) == 1
            assert 'Grandparent' in result
    elif instance.__class__.__name__ == 'GrandChild':
        if exclude_self:
            assert len(result) == 4
            assert 'GrandChild' not in result
        else:
            assert len(result) == 5
            assert 'GrandChild' in result
        assert 'Child' in result
        assert 'ParentA' in result
        assert 'ParentB' in result
        assert 'Grandparent' in result



"""Convenience module that imports all classes from the entire library."""

from highcharts_core.chart import *
from highcharts_core.options import *
from highcharts_core.headless_export import *
from highcharts_core.global_options import *
from highcharts_core.global_options.language import *
from highcharts_core.global_options.shared_options import *
from highcharts_core.global_options.language.accessibility import *
from highcharts_core.options.accessibility import *
from highcharts_core.options.annotations import *
from highcharts_core.options.axes import *
from highcharts_core.options.chart import *
from highcharts_core.options.exporting import *
from highcharts_core.options.legend import *
from highcharts_core.options.navigation import *
from highcharts_core.options.plot_options import *
from highcharts_core.options.series import *
from highcharts_core.options.series.data import *

from highcharts_core.options.series.arcdiagram import ArcDiagramSeries
from highcharts_core.options.series.area import AreaSeries
from highcharts_core.options.series.area import AreaRangeSeries
from highcharts_core.options.series.area import AreaSplineSeries
from highcharts_core.options.series.area import AreaSplineRangeSeries
from highcharts_core.options.series.bar import BarSeries
from highcharts_core.options.series.bellcurve import BellCurveSeries
from highcharts_core.options.series.boxplot import BoxPlotSeries
from highcharts_core.options.series.bubble import BubbleSeries
from highcharts_core.options.series.bullet import BulletSeries
from highcharts_core.options.series.bar import ColumnSeries
from highcharts_core.options.series.bar import ColumnPyramidSeries
from highcharts_core.options.series.bar import ColumnRangeSeries
from highcharts_core.options.series.bar import CylinderSeries
from highcharts_core.options.series.dependencywheel import DependencyWheelSeries
from highcharts_core.options.series.dumbbell import DumbbellSeries
from highcharts_core.options.series.boxplot import ErrorBarSeries
from highcharts_core.options.series.funnel import FunnelSeries
from highcharts_core.options.series.funnel import Funnel3DSeries
from highcharts_core.options.series.gauge import GaugeSeries
from highcharts_core.options.series.heatmap import HeatmapSeries
from highcharts_core.options.series.histogram import HistogramSeries
from highcharts_core.options.series.item import ItemSeries
from highcharts_core.options.series.area import LineSeries
from highcharts_core.options.series.dumbbell import LollipopSeries
from highcharts_core.options.series.networkgraph import NetworkGraphSeries
from highcharts_core.options.series.organization import OrganizationSeries
from highcharts_core.options.series.packedbubble import PackedBubbleSeries
from highcharts_core.options.series.pareto import ParetoSeries
from highcharts_core.options.series.pie import PieSeries
from highcharts_core.options.series.polygon import PolygonSeries
from highcharts_core.options.series.pyramid import PyramidSeries
from highcharts_core.options.series.pyramid import Pyramid3DSeries
from highcharts_core.options.series.sankey import SankeySeries
from highcharts_core.options.series.scatter import ScatterSeries
from highcharts_core.options.series.scatter import Scatter3DSeries
from highcharts_core.options.series.gauge import SolidGaugeSeries
from highcharts_core.options.series.spline import SplineSeries
from highcharts_core.options.series.area import StreamGraphSeries
from highcharts_core.options.series.sunburst import SunburstSeries
from highcharts_core.options.series.heatmap import TilemapSeries
from highcharts_core.options.series.timeline import TimelineSeries
from highcharts_core.options.series.treemap import TreemapSeries
from highcharts_core.options.series.pie import VariablePieSeries
from highcharts_core.options.series.bar import VariwideSeries
from highcharts_core.options.series.vector import VectorSeries
from highcharts_core.options.series.venn import VennSeries
from highcharts_core.options.series.bar import WaterfallSeries
from highcharts_core.options.series.bar import WindBarbSeries
from highcharts_core.options.series.wordcloud import WordcloudSeries
from highcharts_core.options.series.bar import XRangeSeries

from highcharts_core.utility_classes import *



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.sonification.track_configurations import (InstrumentTrackConfiguration,
                                                                       SpeechTrackConfiguration,
                                                                       ContextTrackConfiguration)
from highcharts_core.options.sonification.grouping import SonificationGrouping
from highcharts_core.utility_classes.events import SonificationEvents


class SonificationOptions(HighchartsMeta):
    """Options for configuring sonification and audio charts."""
    
    def __init__(self, **kwargs):
        self._after_series_wait = None
        self._default_instrument_options = None
        self._default_speech_options = None
        self._duration = None
        self._enabled = None
        self._events = None
        self._global_context_tracks = None
        self._global_tracks = None
        self._master_volume = None
        self._order = None
        self._point_grouping = None
        self._show_crosshair = None
        self._show_tooltip = None
        self._update_interval = None
        
        self.after_series_wait = kwargs.get('after_series_wait', None)
        self.default_instrument_options = kwargs.get('default_instrument_options', None)
        self.default_speech_options = kwargs.get('default_speech_options', None)
        self.duration = kwargs.get('duration', None)
        self.enabled = kwargs.get('enabled', None)
        self.events = kwargs.get('events', None)
        self.global_context_tracks = kwargs.get('global_context_tracks', None)
        self.global_tracks = kwargs.get('global_tracks', None)
        self.master_volume = kwargs.get('master_volume', None)
        self.order = kwargs.get('order', None)
        self.point_grouping = kwargs.get('point_grouping', None)
        self.show_crosshair = kwargs.get('show_crosshair', None)
        self.show_tooltip = kwargs.get('show_tooltip', None)
        self.update_interval = kwargs.get('update_interval', None)
        
    @property
    def after_series_wait(self) -> Optional[int | float | Decimal]:
        """The time to wait in milliseconds after each data series when playing the visualization's data series 
        in sequence. Defaults to ``700``.
        
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._after_series_wait
    
    @after_series_wait.setter
    def after_series_wait(self, value):
        self._after_series_wait = validators.numeric(value,
                                                     allow_empty = True,
                                                     minimum = 0)
    
    @property
    def default_instrument_options(self) -> Optional[InstrumentTrackConfiguration]:
        """Default sonification options for all instrument tracks.

        .. warning::
        
          If specific options are also set on individual tracks or per-series, this configuration will be *overridden*.
        
        :rtype: :class:`InstrumentTrackConfiguration <highcharts_core.options.sonification.track_configurations.InstrumentTrackConfiguration>`
          or :obj:`None <python:None>`
        """
        return self._default_instrument_options
    
    @default_instrument_options.setter
    @class_sensitive(InstrumentTrackConfiguration)
    def default_instrument_options(self, value):
        self._default_instrument_options = value
        
    @property
    def default_speech_options(self) -> Optional[SpeechTrackConfiguration]:
        """Default sonification options for all speech tracks.
        .. warning::
        
          If specific options are also set on individual tracks or per-series, this configuration will be *overridden*.
        
        :rtype: :class:`SpeechTrackConfiguration <highcharts_core.options.sonification.track_configurations.SpeechTrackConfiguration>`
          or :obj:`None <python:None>`
        """
        return self._default_speech_options
    
    @default_speech_options.setter
    @class_sensitive(SpeechTrackConfiguration)
    def default_speech_options(self, value):
        self._default_speech_options = value
        
    @property
    def duration(self) -> Optional[int | float | Decimal]:
        """The total duration of the sonification, expressed in milliseconds. Defaults to ``6000``.
        
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._duration
    
    @duration.setter
    def duration(self, value):
        self._duration = validators.numeric(value, allow_empty = True, minimum = 0)
        
    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enables sonification functionality on the chart. Defaults to ``True``.
        
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled
    
    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)
            
    @property
    def events(self) -> Optional[SonificationEvents]:
        """Event handlers for sonification.
        
        :rtype: :class:`SonificationEvents <highcharts_core.utility_classes.events.SonificationEvents>` or
          :obj:`None <python:None>`
        """
        return self._events
    
    @events.setter
    @class_sensitive(SonificationEvents)
    def events(self, value):
        self._events = value
        
    @property
    def global_context_tracks(self) -> Optional[ContextTrackConfiguration]:
        """Context tracks to add globally, an array of either instrument tracks, speech tracks, or a mix.
        
        .. note::
        
          Context tracks are not tied to data points, but play at a set interval - either based on ``time`` or on 
          ``prop`` values.
          
        :rtype: :class:`ContextTrackConfiguration <highcharts_core.options.sonification.track_configurations.ContextTrackConfiguration>`
          or :obj:`None <python:None>`
        """
        return self._global_context_tracks
    
    @global_context_tracks.setter
    @class_sensitive(ContextTrackConfiguration)
    def global_context_tracks(self, value):
        self._global_context_tracks = value
        
    @property
    def global_tracks(self) -> Optional[InstrumentTrackConfiguration]:
        """Global tracks to add to every series.
        
        :rtype: :class:`InstrumentTrackConfiguration <highcharts_core.options.sonification.track_configurations.InstrumentTrackConfiguration>`
          or :obj:`None <python:None>`
        """
        return self._global_tracks
    
    @global_tracks.setter
    @class_sensitive(InstrumentTrackConfiguration)
    def global_tracks(self, value):
        self._global_tracks = value

    @property
    def master_volume(self) -> Optional[int | float | Decimal]:
        """The overall/master volume for the sonification, from ``0`` to ``1``. Defaults to ``0.7``.
        
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._master_volume
    
    @master_volume.setter
    def master_volume(self, value):
        self._master_volume = validators.numeric(value, 
                                                 allow_empty = True,
                                                 minimum = 0,
                                                 maximum = 1)
    
    @property
    def order(self) -> Optional[str]:
        """The order in which to play the sonification for data series. Accepts either:
        
          * ``'sequential'`` where the series play individually one after the other or
          * ``'simultaneous'`` where the series play at the same time
          
        Defaults to ``'sequential'``.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._order
    
    @order.setter
    def order(self, value):
        if not value:
            self._order = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['sequential', 'simultaneous']:
                raise errors.HighchartsValueError(f'.order expects either "sequential" or "simultaenous". '
                                                  f'Received: "{value}"')
            
            self._order = value
            
    @property
    def point_grouping(self) -> Optional[SonificationGrouping]:
        """Options for grouping data points together when sonifying. 
        
        This allows for the visual presentation to contain more points than what is being played. 
        
        If not enabled, all visible / uncropped points are played.
        
        :rtype: :class:`SonificationGrouping <highcharts_core.options.sonification.grouping.SonificationGrouping>` or
          :obj:`None <python:None>`
        """
        return self._point_grouping
    
    @point_grouping.setter
    @class_sensitive(SonificationGrouping)
    def point_grouping(self, value):
        self._point_grouping = value
        
    @property
    def show_crosshair(self) -> Optional[bool]:
        """If ``True``, show X and Y crosshairs (if defined on the chart) as the sonification plays. Defaults to 
        ``True``.
        
        .. warning::
        
           If multiple tracks that play at different times try to show crosshairs, it can be glitchy. Therefore,
           it is recommended in those cases to turn this on/off for individual tracks using the ``.show_play_marker`` 
           property.
           
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_crosshair
    
    @show_crosshair.setter
    def show_crosshair(self, value):
        if value is None:
            self._show_crosshair = None
        else:
            self._show_crosshair = bool(value)
            
    @property
    def show_tooltip(self) -> Optional[bool]:
        """If ``True``, show tooltips as the sonification plays. Defaults to ``True``.
        
        .. warning::
        
           If multiple tracks that play at different times try to show tooltips, it can be glitchy. Therefore,
           it is recommended in those cases to turn this on/off for individual tracks using the ``.show_play_marker`` 
           property.
           
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_tooltip
    
    @show_tooltip.setter
    def show_tooltip(self, value):
        if value is None:
            self._show_tooltip = None
        else:
            self._show_tooltip = bool(value)
    
    @property
    def update_interval(self) -> Optional[int | float | Decimal]:
        """The number of milliseconds to wait between each recomputation of the sonification, if the chart updates 
        rapidly. 
        
        .. tip::
        
          This avoids slowing down processes like panning.
          
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._update_interval
    
    @update_interval.setter
    def update_interval(self, value):
        self._update_interval = validators.numeric(value, allow_empty = True, minimum = 0)
        
    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'after_series_wait': as_dict.get('afterSeriesWait', None),
            'default_instrument_options': as_dict.get('defaultInstrumentOptions', None),
            'default_speech_options': as_dict.get('defaultSpeechOptions', None),
            'duration': as_dict.get('duration', None),
            'enabled': as_dict.get('enabled', None),
            'events': as_dict.get('events', None),
            'global_context_tracks': as_dict.get('globalContextTracks', None),
            'global_tracks': as_dict.get('globalTracks', None),
            'master_volume': as_dict.get('masterVolume', None),
            'order': as_dict.get('order', None),
            'point_grouping': as_dict.get('pointGrouping', None),
            'show_crosshair': as_dict.get('showCrosshair', None),
            'show_tooltip': as_dict.get('showTooltip', None),
            'update_interval': as_dict.get('updateInterval', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'afterSeriesWait': self.after_series_wait,
            'defaultInstrumentOptions': self.default_instrument_options,
            'defaultSpeechOptions': self.default_speech_options,
            'duration': self.duration,
            'enabled': self.enabled,
            'events': self.events,
            'globalContextTracks': self.global_context_tracks,
            'globalTracks': self.global_tracks,
            'masterVolume': self.master_volume,
            'order': self.order,
            'pointGrouping': self.point_grouping,
            'showCrosshair': self.show_crosshair,
            'showTooltip': self.show_tooltip,
            'updateInterval': self.update_interval,
        }

        return untrimmed



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.series.base import SeriesBase
from highcharts_core.options.series.data.arcdiagram import ArcDiagramData
from highcharts_core.options.plot_options.arcdiagram import ArcDiagramOptions
from highcharts_core.utility_functions import mro__to_untrimmed_dict
from highcharts_core.utility_classes.nodes import NodeOptions


class ArcDiagramSeries(SeriesBase, ArcDiagramOptions):
    """Arc diagram series is a chart drawing style in which the vertices of the chart
    are positioned along a line on the Euclidean plane and the edges are drawn as a
    semicircle in one of the two half-planes delimited by the line, or as smooth
    curves formed by sequences of semicircles.

    .. figure:: ../../../_static/arcdiagram-example.png
      :alt: Arc Diagram Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._link_weight = None
        self._nodes = None
        self._offset = None

        self.link_weight = kwargs.get('link_weight', None)
        self.nodes = kwargs.get('nodes', None)
        self.offset = kwargs.get('offset', None)

        super().__init__(**kwargs)

    @property
    def data(self) -> Optional[List[ArcDiagramData]]:
        """The collection of data points for the series. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`list <python:list>` of :class:`ArcDiagramData` or
          :obj:`None <python:None>`
        """
        return self._data

    @data.setter
    @class_sensitive(ArcDiagramData, force_iterable = True)
    def data(self, value):
        self._data = value

    @property
    def link_weight(self) -> Optional[int | float | Decimal]:
        """The global link weight. Defaults to :obj:`None <python:None>`.

        If :obj:`None <python:None>`, width is calculated per link, depending on the
        weight value.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._link_weight

    @link_weight.setter
    def link_weight(self, value):
        if value is None:
            self._link_weight = None
        else:
            self._link_weight = validators.numeric(value,
                                                   minimum = 0)

    @property
    def nodes(self) -> Optional[List[NodeOptions]]:
        """A collection of options for configuring individual nodes. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`list <python:list>` of :class:`NodeOptions` or
          :obj:`None <python:None>`
        """
        return self._nodes

    @nodes.setter
    @class_sensitive(NodeOptions, force_iterable = True)
    def nodes(self, value):
        self._nodes = value

    @property
    def offset(self) -> Optional[str]:
        """The offset of an arc diagram nodes column in relation to the plot area,
        expressed as a percentage of the plot area. If :obj:`None <python:None>`, the
        series is placed so that the biggest node is touching the bottom border of the
        plot area (equivalent to ``'100%'``). Defaults to :obj:`None <python:None>`.

        .. hint::

          For example, setting the offset to ``'50%'`` will place the nodes in the center
          of the chart.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._offset

    @offset.setter
    def offset(self, value):
        if not value:
            self._offset = None
        else:
            value = validators.string(value)
            value = value.lower()
            if '%' not in value:
                raise errors.HighchartsValueError(f'offset expects a percentage string. '
                                                  f'No "%" character found in: {value}')
            self._offset = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'centered_links': as_dict.get('centeredLinks', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'color_index': as_dict.get('colorIndex', None),
            'colors': as_dict.get('colors', None),
            'equal_nodes': as_dict.get('equalNodes', None),
            'levels': as_dict.get('levels', None),
            'link_opacity': as_dict.get('linkOpacity', None),
            'min_link_width': as_dict.get('minLinkWidth', None),
            'node_width': as_dict.get('nodeWidth', None),
            'reversed': as_dict.get('reversed', None),

            'data': as_dict.get('data', None),
            'id': as_dict.get('id', None),
            'index': as_dict.get('index', None),
            'legend_index': as_dict.get('legendIndex', None),
            'name': as_dict.get('name', None),
            'stack': as_dict.get('stack', None),
            'x_axis': as_dict.get('xAxis', None),
            'y_axis': as_dict.get('yAxis', None),
            'z_index': as_dict.get('zIndex', None),

            'link_weight': as_dict.get('linkWeight', None),
            'nodes': as_dict.get('nodes', None),
            'offset': as_dict.get('offset', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'data': self.data,
            'linkWeight': self.link_weight,
            'nodes': self.nodes,
            'offset': self.offset
        }
        parent_as_dict = mro__to_untrimmed_dict(self, in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional
from collections import UserDict
try:
    import orjson as json
except ImportError:
    try:
        import rapidjson as json
    except ImportError:
        try:
            import simplejson as json
        except ImportError:
            import json

import esprima
from esprima.error_handler import Error as ParseError

from validator_collection import validators, errors as validator_errors

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta, JavaScriptDict
from highcharts_core.utility_classes.javascript_functions import CallbackFunction
from highcharts_core.js_literal_functions import assemble_js_literal, serialize_to_js_literal


class MenuItem(HighchartsMeta):
    """Configuration for an item that appears in a context menu."""

    def __init__(self, **kwargs):
        self._onclick = None
        self._text = None
        self._text_key = None
        self._separator = False

        self.onclick = kwargs.get('onclick', None)
        self.text = kwargs.get('text', None)
        self.text_key = kwargs.get('text_key', None)
        self.separator = kwargs.get('separator', None)

    @property
    def onclick(self) -> Optional[CallbackFunction]:
        """JavaScript event callback function which fires when the menu item is clicked.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._onclick

    @onclick.setter
    @class_sensitive(CallbackFunction)
    def onclick(self, value):
        self._onclick = value

    @property
    def text(self) -> Optional[str]:
        """The text to display for the menu item.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._text

    @text.setter
    def text(self, value):
        self._text = validators.string(value, allow_empty = True)

    @property
    def text_key(self) -> Optional[str]:
        """If internationalization is required, the key to a language string.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._text_key

    @text_key.setter
    def text_key(self, value):
        self._text_key = validators.string(value, allow_empty = True)

    @property
    def separator(self) -> Optional[bool]:
        """If ``True``, indicates that the item should be rendered as a separator.
        Defaults to ``False``, but may also be :obj:`None <python:None>`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._separator

    @separator.setter
    def separator(self, value):
        if value is False or value is None:
            self._separator = value
        elif value:
            self._separator = True
        else:
            self._separator = False

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'onclick': as_dict.get('onclick', None),
            'text': as_dict.get('text', None),
            'text_key': as_dict.get('textKey', None),
            'separator': as_dict.get('separator', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'onclick': self.onclick,
            'text': self.text,
            'textKey': self.text_key,
            'separator': self.separator
        }

        return untrimmed


class MenuObject(JavaScriptDict):
    """Special :class:`dict <python:dict>` class used to construct a Highcharts menu
    configuration. Each key represents the identifier of a menu item, while the object
    is a configuration of that menu item's settings.

    Keys are validated to be valid variable names, while each value is validated to be a
    :class:`MenuItem`.

    """
    _valid_value_types = MenuItem
    _allow_empty_value = False



from typing import Optional, List

from highcharts_core.options.series.base import SeriesBase
from highcharts_core.options.series.data.venn import VennData
from highcharts_core.options.plot_options.venn import VennOptions
from highcharts_core.utility_functions import mro__to_untrimmed_dict


class VennSeries(SeriesBase, VennOptions):
    """Options to configure a Venn series.

    A Venn diagram displays all possible logical relations between a collection of
    different sets. The sets are represented by circles, and the relation between the
    sets are displayed by the overlap or lack of overlap between them. The venn
    diagram is a special case of Euler diagrams, which can also be displayed by this
    series type.

    .. tabs::

      .. tab:: Venn Diagram

        .. figure:: ../../../_static/venn-example.png
          :alt: Venn Example Chart
          :align: center

      .. tab:: Euler Diagram

        .. figure:: ../../../_static/venn-example-euler.png
          :alt: Euler Example Chart
          :align: center

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def data(self) -> Optional[List[VennData]]:
        """Collection of data that represents the series. Defaults to
        :obj:`None <python:None>`.

        While the series type returns a collection of :class:`VennData` instances,
        it accepts as input three different types of data:

        .. tabs::

          .. tab:: Object Collection

            A one-dimensional collection of :class:`VennData` objects or objects
            coercable to :class:`VennData`.

        :rtype: :class:`list <python:list>` of :class:`VennData` or
          :obj:`None <python:None>`
        """
        return self._data

    @data.setter
    def data(self, value):
        if not value:
            self._data = None
        else:
            self._data = VennData.from_array(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'point_description_format': as_dict.get('pointDescriptionFormat', None),
            'point_interval': as_dict.get('pointInterval', None),            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'data': as_dict.get('data', None),
            'id': as_dict.get('id', None),
            'index': as_dict.get('index', None),
            'legend_index': as_dict.get('legendIndex', None),
            'name': as_dict.get('name', None),
            'stack': as_dict.get('stack', None),
            'x_axis': as_dict.get('xAxis', None),
            'y_axis': as_dict.get('yAxis', None),
            'z_index': as_dict.get('zIndex', None),

            'border_dash_style': as_dict.get('borderDashStyle', None),
            'brighten': as_dict.get('brighten', None),
            'cluster': as_dict.get('cluster', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = mro__to_untrimmed_dict(self, in_cls = in_cls)

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class AnnounceNewDataLanguageOptions(HighchartsMeta):
    """Default announcement for new data in charts.

    .. note::

      If (JavaScript) ``addPoint()`` or ``addSeries()`` is used, and only one
      series/point is added, the
      :meth:`AnnounceNewDataLanguageOptions.new_point_announce` and
      :meth:`AnnounceNewdata.new_series_announce` strings are used.

      The ``...single`` versions will be used if there is only one chart on the page,
      and the ``...multiple`` versions will be used if there are multiple charts on the
      page. For all other new data events, the
      :meth:`AnnounceNewData.new_data_announce` string will be used.

    """

    def __init__(self, **kwargs):
        self._new_data_announce = None
        self._new_point_announce_multiple = None
        self._new_point_announce_single = None
        self._new_series_announce_multiple = None
        self._new_series_announce_single = None

        self.new_data_announce = kwargs.get('new_data_announce', None)
        self.new_point_announce_multiple = kwargs.get('new_point_announce_multiple', None)
        self.new_point_announce_single = kwargs.get('new_point_announce_single', None)
        self.new_series_announce_multiple = kwargs.get('new_series_announce_multiple',
                                                       None)
        self.new_series_announce_single = kwargs.get('new_series_announce_single', None)

    @property
    def new_data_announce(self) -> Optional[str]:
        """Announcement for any new data. Defaults to:
        ``'Updated data for chart {chartTitle}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._new_data_announce

    @new_data_announce.setter
    def new_data_announce(self, value):
        self._new_data_announce = validators.string(value, allow_empty = True)

    @property
    def new_point_announce_multiple(self) -> Optional[str]:
        """Announcement when multiple new points have been added. Defaults to:
        ``'New data point in chart {chartTitle}: {pointDesc}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._new_point_announce_multiple

    @new_point_announce_multiple.setter
    def new_point_announce_multiple(self, value):
        self._new_point_announce_multiple = validators.string(value, allow_empty = True)

    @property
    def new_point_announce_single(self) -> Optional[str]:
        """Announcement when a single new point has been added. Defaults to:
        ``'New data point: {pointDesc}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._new_point_announce_single

    @new_point_announce_single.setter
    def new_point_announce_single(self, value):
        self._new_point_announce_single = validators.string(value, allow_empty = True)

    @property
    def new_series_announce_multiple(self) -> Optional[str]:
        """Announcement when multiple new series have been added. Defaults to:
        ``'New data series in chart {chartTitle}: {seriesDesc}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._new_series_announce_multiple

    @new_series_announce_multiple.setter
    def new_series_announce_multiple(self, value):
        self._new_series_announce_multiple = validators.string(value, allow_empty = True)

    @property
    def new_series_announce_single(self) -> Optional[str]:
        """Announcement when a single new series has been added. Defaults to:
        ``'New data series: {seriesDesc}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._new_series_announce_single

    @new_series_announce_single.setter
    def new_series_announce_single(self, value):
        self._new_series_announce_single = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'new_data_announce': as_dict.get('newDataAnnounce', None),
            'new_point_announce_multiple': as_dict.get('newPointAnnounceMultiple', None),
            'new_point_announce_single': as_dict.get('newPointAnnounceSingle', None),
            'new_series_announce_multiple': as_dict.get('newSeriesAnnounceMultiple',
                                                        None),
            'new_series_announce_single': as_dict.get('newSeriesAnnounceSingle', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'newDataAnnounce': self.new_data_announce,
            'newPointAnnounceMultiple': self.new_point_announce_multiple,
            'newPointAnnounceSingle': self.new_point_announce_single,
            'newSeriesAnnounceMultiple': self.new_series_announce_multiple,
            'newSeriesAnnounceSingle': self.new_series_announce_single
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.global_options.language import Language as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'context_button_title': 'some title goes here',
      'decimal_point': '--',
      'download_csv': 'Download CSV',
      'download_jpeg': 'Download JPEG',
      'download_pdf': 'Download PDF',
      'download_png': 'Download PNG',
      'download_svg': 'Download SVG',
      'download_xls': 'Download Excel',
      'drillup_text': 'Backup',
      'exit_fullscreen': 'Exit Fullscreen Mode',
      'hide_data': 'Hide',
      'invalid_date': 'Invalid Time Period',
      'loading': 'Loading...',
      'main_breadcrumb': 'Home',
      'months': [
          'Jan',
          'Feb',
          'Mar',
          'Apr',
          'May',
          'Jun',
          'Jul',
          'Aug',
          'Sep',
          'Oct',
          'Nov',
          'Dec'
      ],
      'no_data': 'No data available.',
      'numeric_symbol_magnitude': 1,
      'numeric_symbols': ['#', 'k', 'm'],
      'print_chart': 'Print',
      'reset_zoom': 'Reset',
      'reset_zoom_title': 'Reset Zoom',
      'short_months': [
          'Jan',
          'Feb',
          'Mar',
          'Apr',
          'May',
          'Jun',
          'Jul',
          'Aug',
          'Sep',
          'Oct',
          'Nov',
          'Dec'
      ],
      'short_weekdays': [
          'Mon',
          'Tue',
          'Wed',
          'Thurs',
          'Fri',
          'Sat',
          'Sun'
      ],
      'thousands_separator': ',',
      'view_data': 'View Data',
      'view_fullscreen': 'Fullscreen',
      'weekdays': [
          'Mon',
          'Tue',
          'Wed',
          'Thurs',
          'Fri',
          'Sat',
          'Sun'
      ]
    }, None),
    ({
      'export_data': {
        'annotationHeader': 'some string',
        'categoryDatetimeHeader': 'some string',
        'categoryHeader': 'a different string'
      }
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('global_options/language/language/01.js', False, None),

    ('global_options/language/language/error-01.js', False, (errors.HighchartsValueError,
                                                             errors.HighchartsParseError,
                                                             JSONDecodeError,
                                                             TypeError,
                                                             ValueError)),

    ('global_options/language/language/01.js', True, None),

    ('global_options/language/language/error-01.js', True, (errors.HighchartsValueError,
                                                            errors.HighchartsParseError,
                                                            JSONDecodeError,
                                                            TypeError,
                                                            ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from copy import deepcopy
from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.options.plot_options.pie import PieOptions
from highcharts_core.options.plot_options.bar import ColumnOptions


class FunnelOptions(PieOptions):
    """General options to apply to all Funnel series types.

    Funnel charts are a type of chart often used to visualize stages in a sales
    project, where the top are the initial stages with the most clients.

    .. warning::

      Funnel charts require that the ``modules/funnel.js`` file is loaded client-side.

    .. figure:: ../../../_static/funnel-example.png
      :alt: Funnel Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._border_radius = None
        self._height = None
        self._neck_height = None
        self._neck_width = None
        self._reversed = None
        self._width = None

        self.border_radius = kwargs.get('border_radius', None)
        self.height = kwargs.get('height', None)
        self.neck_height = kwargs.get('neck_height', None)
        self.neck_width = kwargs.get('neck_width', None)
        self.reversed = kwargs.get('reversed', None)
        self.width = kwargs.get('width', None)

        super().__init__(**kwargs)

    @property
    def border_radius(self) -> Optional[str | int | float | Decimal]:
        """The corner radius of the border surrounding all points or series. Defaults to ``0``.
        
        .. note::
        
          A number signifies pixels. A percentage string (e.g.``'50%'``) signifies a size 
          relative to the series width.
          
        :rtype: :class:`str <python:str>`, numeric, or :obj:`None <python:None>`
        """
        return self._border_radius
        
    @border_radius.setter
    def border_radius(self, value):
        if value is None:
            self._border_radius = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('border_radius expects a number or a "%" '
                                                      'string. No "%" character found.')
            except TypeError:
                value = validators.numeric(value, minimum = 0)

            self._border_radius = value
        

    @property
    def height(self) -> Optional[str | int | float | Decimal]:
        """The height of the funnel or pyramid. Defaults to ``'100%'``.

        If it is a number, it defines the height in pixels. If it is a percentage string,
        the height represents the percentage of the plot area height.

        :rtype: :class:`str <python:str>`, numeric, or :obj:`None <python:None>`
        """
        return self._height

    @height.setter
    def height(self, value):
        if value is None:
            self._height = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('height expects a number or a "%" '
                                                      'string. No "%" character found.')
            except TypeError:
                value = validators.numeric(value, minimum = 0)

            self._height = value

    @property
    def neck_height(self) -> Optional[str | int | float | Decimal]:
        """The height of the neck, the lower part of the funnel. Defaults to ``'25%'``.

        A number defines pixel width, a percentage string defines a percentage of the
        plot area height.

        :rtype: :class:`str <python:str>`, numeric, or :obj:`None <python:None>`
        """
        return self._neck_height

    @neck_height.setter
    def neck_height(self, value):
        if value is None:
            self._neck_height = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('neck_height expects a number or a '
                                                      '"%" string. No "%" character '
                                                      'found.')
            except TypeError:
                value = validators.numeric(value, minimum = 0)

            self._neck_height = value

    @property
    def neck_width(self) -> Optional[str | int | float | Decimal]:
        """The width of the neck, the lower part of the funnel. Defaults to ``'30%'``.

        A number defines pixel width, a percentage string defines a percentage of the
        plot area width.

        :rtype: :class:`str <python:str>`, numeric, or :obj:`None <python:None>`
        """
        return self._neck_width

    @neck_width.setter
    def neck_width(self, value):
        if value is None:
            self._neck_width = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('neck_width expects a number or a '
                                                      '"%" string. No "%" character '
                                                      'found.')
            except TypeError:
                value = validators.numeric(value, minimum = 0)

            self._neck_width = value

    @property
    def reversed(self) -> Optional[bool]:
        """If ``True``, the widest area of the funnel is lower down. A reversed funnel
        with no neck width and neck height is a pyramid.

        Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._reversed

    @reversed.setter
    def reversed(self, value):
        if value is None:
            self._reversed = None
        else:
            self._reversed = bool(value)

    @property
    def width(self) -> Optional[str | int | float | Decimal]:
        """The width of the funnel. Defaults to ``'90%'``.

        A number defines pixel width, a percentage string defines a percentage of the
        plot area width.

        :rtype: :class:`str <python:str>`, numeric, or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        if value is None:
            self._width = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('width expects a number or a '
                                                      '"%" string. No "%" character '
                                                      'found.')
            except TypeError:
                value = validators.numeric(value, minimum = 0)

            self._width = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'center': as_dict.get('center', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'colors': as_dict.get('colors', None),
            'depth': as_dict.get('depth', None),
            'end_angle': as_dict.get('endAngle', None),
            'fill_color': as_dict.get('fillColor', None),
            'ignore_hidden_point': as_dict.get('ignoreHiddenPoint', None),
            'inner_size': as_dict.get('innerSize', None),
            'linecap': as_dict.get('linecap', None),
            'min_size': as_dict.get('minSize', None),
            'size': as_dict.get('size', None),
            'sliced_offset': as_dict.get('slicedOffset', None),
            'start_angle': as_dict.get('startAngle', None),
            'thickness': as_dict.get('thickness', None),

            'border_radius': as_dict.get('borderRadius', None),
            'height': as_dict.get('height', None),
            'neck_height': as_dict.get('neckHeight', None),
            'neck_width': as_dict.get('neckWidth', None),
            'reversed': as_dict.get('reversed', None),
            'width': as_dict.get('width', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderRadius': self.border_radius,
            'height': self.height,
            'neckHeight': self.neck_height,
            'neckWidth': self.neck_width,
            'reversed': self.reversed,
            'width': self.width
        }
        parent_as_dict = self._untrimmed_mro_ancestors(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class Funnel3DOptions(FunnelOptions, ColumnOptions):
    """General options to apply to all Funnel 3D series types.

    A Funnel 3D chart is a three-dimensional version of funnel series type. Funnel
    charts are a type of chart often used to visualize stages in a sales project,
    where the top are the initial stages with the most clients.

    .. warning::

      Funnel 3D charts require that the following files are all loaded client-side:

        * ``highcharts-3d.js``,
        * ``cylinder.js`` and
        * ``funnel3d.js``

    .. figure:: ../../../_static/funnel_3d-example.png
      :alt: Funnel 3D Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._gradient_for_sides = None

        self.gradient_for_sides = kwargs.get('gradient_for_sides', None)

        super().__init__(**kwargs)

    @property
    def gradient_for_sides(self) -> Optional[bool]:
        """If ``True``, sides fill is set to a gradient. If ``False``, sides are rendered
        with a solid color. Defaults to :obj:`None <python:None>`, which is equivalent to
        ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`.
        """
        return self._gradient_for_sides

    @gradient_for_sides.setter
    def gradient_for_sides(self, value):
        if value is None:
            self._gradient_for_sides = None
        else:
            self._gradient_for_sides = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'gradient_for_sides': as_dict.get('gradientForSides', None)
        }
        mro = cls.mro()
        parent_kwargs = [x._get_kwargs_from_dict(as_dict) for x in mro[1:2]]
        for kwarg_set in parent_kwargs:
            for key in kwarg_set:
                kwargs[key] = kwarg_set[key]

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'gradientForSides': self.gradient_for_sides,
        }
        parent_as_dict = self._untrimmed_mro_ancestors(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.global_options.shared_options import SharedOptions as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal, Class_from_js_literal_with_expected

STANDARD_PARAMS = [
    ({}, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('input_filename, expected_filename, as_file, error', [
    ('global_options/shared_options/01-input.js',
     'global_options/shared_options/01-expected.js',
     False,
     None),
    ('global_options/shared_options/01-input.js',
     'global_options/shared_options/01-expected.js',
     False,
     None),
])
def test_from_js_literal(input_files, input_filename, expected_filename, as_file, error):
    Class_from_js_literal_with_expected(cls,
                                        input_files,
                                        input_filename,
                                        expected_filename,
                                        as_file,
                                        error)



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class ScreenReaderSection(HighchartsMeta):
    """Accessibility options for the screen reader information sections added before
    and after the chart."""

    def __init__(self, **kwargs):
        self._after_chart_format = None
        self._after_chart_formatter = None
        self._axis_range_date_format = None
        self._before_chart_format = None
        self._before_chart_formatter = None
        self._on_play_as_sound_click = None
        self._on_view_data_table_click = None

        self.after_chart_format = kwargs.get('after_chart_format', None)
        self.after_chart_formatter = kwargs.get('after_chart_formatter', None)
        self.axis_range_date_format = kwargs.get('axis_range_date_format', None)
        self.before_chart_format = kwargs.get('before_chart_format', None)
        self.before_chart_formatter = kwargs.get('before_chart_formatter', None)
        self.on_play_as_sound_click = kwargs.get('on_play_as_sound_click', None)
        self.on_view_data_table_click = kwargs.get('on_view_data_table_click', None)

    @property
    def after_chart_format(self) -> Optional[str]:
        """Format for the screen reader information region after the chart. Defaults to
        ``'{endOfChartMarker}'``.

        Supported HTML tags are:
          * ``<h1-6>``
          * ``<p>``
          * ``<div>``
          * ``<a>``
          * ``<ul>``
          * ``<ol>``
          * ``<li>``
          * ``<button>``

        Attributes are not supported, except for ``id`` on ``<div>``, ``<a>``, and
        ``<button>``.

        ``id`` is required on ``<a>`` and ``<button>`` in the format
        ``<tag id="abcd">``. Numbers, lower- and uppercase letters, ``"-"`` and ``"#"``
        are valid characters in IDs.

        The ``headingTagName`` is an auto-detected heading (``h1-h6``) that corresponds to
        the heading level below the previous heading in the DOM.

        .. tip::

          Set to empty string to remove the region altogether.

        :returns: Content to render in the screen reader information region after the
          chart.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._after_chart_format

    @after_chart_format.setter
    def after_chart_format(self, value):
        if value is None:
            self._after_chart_format = None
        elif not value:
            self._after_chart_format = ''
        else:
            self._after_chart_format = validators.string(value, allow_empty = False)

    @property
    def after_chart_formatter(self) -> Optional[CallbackFunction]:
        """A JavaScript formatter function to create the HTML contents of the hidden
        screen reader information region after the chart.

        The formatter function should receive one argument, ``chart``, referring to the
        chart object. It should return a string with the HTML content of the region.

        If :obj:`None <python:None>`, will returns an automatic description of the chart
        based on :meth:`ScreenReaderSection.after_chart_format`.

        :returns: JavaScript formatter function for the screen reader information region
          after the chart.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._after_chart_formatter

    @after_chart_formatter.setter
    @class_sensitive(CallbackFunction)
    def after_chart_formatter(self, value):
        self._after_chart_formatter = value

    @property
    def axis_range_date_format(self) -> Optional[str]:
        """Date format to use to describe range of datetime axes. Defaults to
        ``%Y-%m-%d %H:%M:%S``.

        .. seealso::

          * Detailed documentation on supported format replacement codes:
            https://api.highcharts.com/class-reference/Highcharts.Time#dateFormat

        :returns: Date format to use to describe range of datetime axes.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._axis_range_date_format

    @axis_range_date_format.setter
    def axis_range_date_format(self, value):
        self._axis_range_date_format = validators.string(value, allow_empty = True)

    @property
    def before_chart_format(self) -> Optional[str]:
        """Format for the screen reader information region before the chart. Defaults to

        .. code-block::

          '<{headingTagName}>{chartTitle}</{headingTagName}>
           <div>{typeDescription}</div><div>{chartSubtitle}</div>
           <div>{chartLongdesc}</div><div>{playAsSoundButton}</div>
           <div>{viewTableButton}</div><div>{xAxisDescription}</div>
           <div>{yAxisDescription}</div>
           <div>{annotationsTitle}{annotationsList}</div>'

        Supported HTML tags are:
          * ``<h1-6>``
          * ``<p>``
          * ``<div>``
          * ``<a>``
          * ``<ul>``
          * ``<ol>``
          * ``<li>``
          * ``<button>``

        Attributes are not supported, except for ``id`` on ``<div>``, ``<a>``, and
        ``<button>``.

        ``id`` is required on ``<a>`` and ``<button>`` in the format
        ``<tag id="abcd">``. Numbers, lower- and uppercase letters, ``"-"`` and ``"#"``
        are valid characters in IDs.

        The ``headingTagName`` is an auto-detected heading (``h1-h6``) that corresponds to
        the heading level below the previous heading in the DOM.

        .. tip::

          Set to empty string to remove the region altogether.

        :returns: Content to render in the screen reader information region before the
          chart.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._before_chart_format

    @before_chart_format.setter
    def before_chart_format(self, value):
        if value is None:
            self._before_chart_format = None
        elif not value:
            self._before_chart_format = ''
        else:
            self._before_chart_format = validators.string(value, allow_empty = False)

    @property
    def before_chart_formatter(self) -> Optional[CallbackFunction]:
        """A JavaScript formatter function to create the HTML contents of the hidden
        screen reader information region before the chart.

        The formatter function should receive one argument, ``chart``, referring to the
        chart object. It should return a string with the HTML content of the region.

        If :obj:`None <python:None>`, will returns an automatic description of the chart
        based on :meth:`ScreenReaderSection.before_chart_format`.

        :returns: JavaScript formatter function for the screen reader information region
          before the chart.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._before_chart_formatter

    @before_chart_formatter.setter
    @class_sensitive(CallbackFunction)
    def before_chart_formatter(self, value):
        self._before_chart_formatter = value

    @property
    def on_play_as_sound_click(self) -> Optional[CallbackFunction]:
        """JavaScript function to run upon clicking the "Play as sound" button in the
        screen reader region.

        By default Highcharts will call the ``chart.sonify`` JavaScript function.

        :returns: JavaScript function to run upon clicking the "Play as sound" button in
          the screen reader region.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._on_play_as_sound_click

    @on_play_as_sound_click.setter
    @class_sensitive(CallbackFunction)
    def on_play_as_sound_click(self, value):
        self._on_play_as_sound_click = value

    @property
    def on_view_data_table_click(self) -> Optional[CallbackFunction]:
        """JavaScript function to run upon clicking the "View as Data Table" link in the
        screen reader region.

        By default Highcharts will insert and set focus to a data table representation of
        the chart.

        :returns: JavaScript function to run upon clicking the "View as Data Table" link
          in the screen reader region.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._on_view_data_table_click

    @on_view_data_table_click.setter
    @class_sensitive(CallbackFunction)
    def on_view_data_table_click(self, value):
        self._on_view_data_table_click = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'after_chart_format': as_dict.get('afterChartFormat', None),
            'after_chart_formatter': as_dict.get('afterChartFormatter', None),
            'axis_range_date_format': as_dict.get('axisRangeDateFormat', None),
            'before_chart_format': as_dict.get('beforeChartFormat', None),
            'before_chart_formatter': as_dict.get('beforeChartFormatter', None),
            'on_play_as_sound_click': as_dict.get('onPlayAsSoundClick', None),
            'on_view_data_table_click': as_dict.get('onViewDataTableClick', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'afterChartFormat': self.after_chart_format,
            'afterChartFormatter': self.after_chart_formatter,
            'axisRangeDateFormat': self.axis_range_date_format,
            'beforeChartFormat': self.before_chart_format,
            'beforeChartFormatter': self.before_chart_formatter,
            'onPlayAsSoundClick': self.on_play_as_sound_click,
            'onViewDataTableClick': self.on_view_data_table_click
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core.metaclasses import HighchartsMeta


class PDFFontOptions(HighchartsMeta):
    """Settings for a custom font for the exported PDF, when using the
    ``offline-exporting`` module.

    This is used for languages containing non-ASCII characters, like Chinese, Russian,
    Japanese etc.

    As described in the
    `jsPDF docs <https://github.com/parallax/jsPDF#use-of-unicode-characters--utf-8>`_,
    the 14 standard fonts in PDF are limited to the ASCII-codepage. Therefore, in
    order to support other text in the exported PDF, one or more TTF font files have
    to be passed on to the exporting module.

    """

    def __init__(self, **kwargs):
        self._bold = None
        self._bolditalic = None
        self._italic = None
        self._normal = None

        self.bold = kwargs.get('bold', None)
        self.bolditalic = kwargs.get('bolditalic', None)
        self.italic = kwargs.get('italic', None)
        self.normal = kwargs.get('normal', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'exporting.pdfFont'

    @property
    def bold(self) -> Optional[str]:
        """The TTF font file for bold text. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises ValueError: if not a URL or path-like string
        """
        return self._bold

    @bold.setter
    def bold(self, value):
        try:
            self._bold = validators.url(value, allow_empty = True)
        except ValueError:
            self._bold = validators.path(value, allow_empty = True)

    @property
    def bolditalic(self) -> Optional[str]:
        """The TTF font file for Bold Italic text. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises ValueError: if not a URL or path-like string
        """
        return self._bolditalic

    @bolditalic.setter
    def bolditalic(self, value):
        try:
            self._bolditalic = validators.url(value, allow_empty = True)
        except ValueError:
            self._bolditalic = validators.path(value, allow_empty = True)

    @property
    def italic(self) -> Optional[str]:
        """The TTF font file for italic text. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises ValueError: if not a URL or path-like string
        """
        return self._italic

    @italic.setter
    def italic(self, value):
        try:
            self._italic = validators.url(value, allow_empty = True)
        except ValueError:
            self._italic = validators.path(value, allow_empty = True)

    @property
    def normal(self) -> Optional[str]:
        """The TTF font file for normal ``font-style``. Defaults to
        :obj:`None <python:None>`.

        .. note::

           If font variations like bold or italic are not defined, the normal font
           will be used for those, too.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises ValueError: if not a URL or path-like string
        """
        return self._normal

    @normal.setter
    def normal(self, value):
        try:
            self._normal = validators.url(value, allow_empty = True)
        except ValueError:
            self._normal = validators.path(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'bold': as_dict.get('bold', None),
            'bolditalic': as_dict.get('bolditalic', None),
            'italic': as_dict.get('italic', None),
            'normal': as_dict.get('normal', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'bold': self.bold,
            'bolditalic': self.bolditalic,
            'italic': self.italic,
            'normal': self.normal
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.plot_options.series import SeriesOptions
from highcharts_core.utility_classes.clusters import ClusterOptions


class VectorOptions(SeriesOptions):
    """General options to apply to all Vector series types.

    A vector plot is a type of cartesian chart where each point has an X and Y
    position, a length and a direction. Vectors are drawn as arrows.

    .. figure:: ../../../_static/vector-example.png
      :alt: Vector Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._cluster = None
        self._rotation_origin = None
        self._vector_length = None

        self.cluster = kwargs.get('cluster', None)
        self.rotation_origin = kwargs.get('rotation_origin', None)
        self.vector_length = kwargs.get('vector_length', None)

        super().__init__(**kwargs)

    @property
    def cluster(self) -> Optional[ClusterOptions]:
        """Options for marker clusters, the concept of sampling the data values into
        larger blocks in order to ease readability and increase performance of the
        JavaScript charts.

        .. warning::

          The marker clusters module does not work with ``boost`` and ``draggable-points``
          modules.

        .. note::

          The marker clusters feature requires the ``marker-clusters.js`` file to be
          loaded, found in the modules directory of the download package, or online at
          `code.highcharts.com/modules/marker-clusters.js <code.highcharts.com/modules/marker-clusters.js>`_.

        :rtype: :class:`ClusterOptions` or :obj:`None <python:None>`
        """
        return self._cluster

    @cluster.setter
    @class_sensitive(ClusterOptions)
    def cluster(self, value):
        self._cluster = value

    @property
    def rotation_origin(self) -> Optional[str]:
        """What part of the vector it should be rotated around. Defaults to ``'center'``.

        Supports the following values:

          * ``'start'``
          * ``'center'``
          * ``'end'``

        If ``'start'``, the vectors will start from the given [x, y] position, and when
        ``'end'`` the vectors will end in the [x, y] position.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._rotation_origin

    @rotation_origin.setter
    def rotation_origin(self, value):
        if not value:
            self._rotation_origin = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['start', 'center', 'end']:
                raise errors.HighchartsValueError(f'rotation_origin expects either '
                                                  f'"start", "center", or "end". '
                                                  f'Received: {value}')

            self._rotation_origin = value

    @property
    def vector_length(self) -> Optional[int | float | Decimal]:
        """Maximum length of the arrows in the vector plot. Defaults to ``20``.

        .. note::

          The individual arrow length is computed between 0 and this value.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._vector_length

    @vector_length.setter
    def vector_length(self, value):
        self._vector_length = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'point_description_format': as_dict.get('pointDescriptionFormat', None),
            'point_interval': as_dict.get('pointInterval', None),            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'cluster': as_dict.get('cluster', None),
            'rotation_origin': as_dict.get('rotationOrigin', None),
            'vector_length': as_dict.get('vectorLength', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'cluster': self.cluster,
            'rotationOrigin': self.rotation_origin,
            'vectorLength': self.vector_length
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



"""Tests for ``highcharts.pane``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.pane import Pane as cls
from highcharts_core.options.pane import PaneBackground
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'background': [{
          'className': 'test-class-name',
          'innerRadius': '24%',
          'shape': 'circle'
      }],
      'center': ['50%', '50%'],
      'size': '120',
      'start_angle': 0
    }, None),
    ({
      'background': [
          {
            'className': 'test-class-name',
            'innerRadius': '24%',
            'shape': 'circle'
          },
          {
            'innerRadius': '35%',
          }
      ],
      'center': ['50%', '50%'],
      'size': '120',
      'start_angle': 0
    }, None),

    ({
        'background': [{
            'className': 'test-class-name',
            'innerRadius': '24%',
            'shape': 'circle'
        }],
        'center': ['not-valid', 'not-valid'],
        'size': '120',
        'start_angle': 0
    }, (ValueError, TypeError)),
    ({
        'background': [
            {
                'className': 'test-class-name',
                'innerRadius': '24%',
                'shape': 'circle'
            },
            {
                'innerRadius': '35',
            }
        ],
        'center': ['not-valid'],
        'size': 120,
        'start_angle': 0
    }, (ValueError, TypeError)),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('pane/01.js', False, None),
    ('pane/02.js', False, None),

    ('pane/error-01.js', False, (errors.HighchartsValueError,
                                 errors.HighchartsParseError,
                                 JSONDecodeError,
                                 ValueError)),
    ('pane/error-02.js', False, (errors.HighchartsValueError,
                                 errors.HighchartsParseError,
                                 JSONDecodeError,
                                 ValueError)),

    ('pane/01.js', True, None),
    ('pane/02.js', True, None),

    ('pane/error-01.js', True, (errors.HighchartsValueError,
                                errors.HighchartsParseError,
                                JSONDecodeError,
                                ValueError)),
    ('pane/error-02.js', True, (errors.HighchartsValueError,
                                errors.HighchartsParseError,
                                JSONDecodeError,
                                ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional, List

from highcharts_core.options.series.base import SeriesBase
from highcharts_core.options.series.data.treegraph import TreegraphData
from highcharts_core.options.plot_options.treegraph import TreegraphOptions
from highcharts_core.utility_functions import mro__to_untrimmed_dict


class TreegraphSeries(SeriesBase, TreegraphOptions):
    """General options to apply to all :term:`Treegraph` series types.
    
    A treegraph visualizes a relationship between ancestors and descendants with a clear parent-child relationship,
    e.g. a family tree or a directory structure.
    
    .. figure:: ../../../_static/treegraph-example.png
      :alt: Treegraph Example Chart
      :align: center
    
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def data(self) -> Optional[List[TreegraphData]]:
        """Collection of data that represents the series. Defaults to
        :obj:`None <python:None>`.

        While the series type returns a collection of :class:`TreegraphData` instances,
        it accepts as input:

        .. tabs::

          .. tab:: 1D Array of Arrays
          
            A one-dimensional collection where each member of the collection is itself
            a collection of data points.
            
            .. note::
            
              If using the Array of Arrays pattern you *must* set 
              :meth:`.keys <highcharts_core.options.series.treegraph.TreegraphSeries.keys>` to indicate
              which value in the inner array corresponds to 
              :meth:`.id <highcharts_core.options.series.treegraph.TreegraphSeries.id>`, 
              :meth:`.parent <highcharts_core.options.series.treegraph.TreegraphSeries.parent>`, or
              :meth:`.name <highcharts_core.options.series.treegraph.TreegraphSeries.name>`.
              
          .. tab:: Object Collection

            A one-dimensional collection of :class:`TreegraphData` objects or
            :class:`dict <python:dict>` instances coercable to :class:`TreegraphData`

        :rtype: :class:`list <python:list>` of :class:`TreegraphData` or
          :obj:`None <python:None>`
        """
        return self._data

    @data.setter
    def data(self, value):
        if not value:
            self._data = None
        else:
            self._data = TreegraphData.from_array(value)
            
    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_index': as_dict.get('colorIndex', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'step': as_dict.get('step', None),

            'point_interval': as_dict.get('pointInterval', None),
            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_start': as_dict.get('pointStart', None),
            'stacking': as_dict.get('stacking', None),

            'allow_traversing_tree': as_dict.get('allowTraversingTree', None),
            'collapse_button': as_dict.get('collapseButton', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'fill_space': as_dict.get('fillSpace', None),
            'link': as_dict.get('link', None),
            'reversed': as_dict.get('reversed', None),

            'data': as_dict.get('data', None),
            'id': as_dict.get('id', None),
            'index': as_dict.get('index', None),
            'legend_index': as_dict.get('legendIndex', None),
            'name': as_dict.get('name', None),

        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = mro__to_untrimmed_dict(self, in_cls = in_cls) or {}

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class AxisLanguageOptions(HighchartsMeta):
    """Axis description strings."""

    def __init__(self, **kwargs):
        self._range_categories = None
        self._range_from_to = None
        self._time_range_days = None
        self._time_range_hours = None
        self._time_range_minutes = None
        self._time_range_seconds = None
        self._x_axis_description_plural = None
        self._x_axis_description_singular = None
        self._y_axis_description_plural = None
        self._y_axis_description_singular = None

        self.range_categories = kwargs.get('range_categories', None)
        self.range_from_to = kwargs.get('range_from_to', None)
        self.time_range_days = kwargs.get('time_range_days', None)
        self.time_range_hours = kwargs.get('time_range_hours', None)
        self.time_range_minutes = kwargs.get('time_range_minutes', None)
        self.time_range_seconds = kwargs.get('time_range_seconds', None)
        self.x_axis_description_plural = kwargs.get('x_axis_description_plural', None)
        self.x_axis_description_singular = kwargs.get('x_axis_description_singular', None)
        self.y_axis_description_plural = kwargs.get('y_axis_description_plural', None)
        self.y_axis_description_singular = kwargs.get('y_axis_description_singular', None)

    @property
    def range_categories(self) -> Optional[str]:
        """Defaults to ``'Data range: {numCategories} categories.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._range_categories

    @range_categories.setter
    def range_categories(self, value):
        self._range_categories = validators.string(value, allow_empty = True)

    @property
    def range_from_to(self) -> Optional[str]:
        """Defaults to ``'Data ranges from {rangeFrom} to {rangeTo}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._range_from_to

    @range_from_to.setter
    def range_from_to(self, value):
        self._range_from_to = validators.string(value, allow_empty = True)

    @property
    def time_range_days(self) -> Optional[str]:
        """Defaults to ``'Data range: {range} days.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._time_range_days

    @time_range_days.setter
    def time_range_days(self, value):
        self._time_range_days = validators.string(value, allow_empty = True)

    @property
    def time_range_hours(self) -> Optional[str]:
        """Defaults to ``'Data range: {range} hours.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._time_range_hours

    @time_range_hours.setter
    def time_range_hours(self, value):
        self._time_range_hours = validators.string(value, allow_empty = True)

    @property
    def time_range_minutes(self) -> Optional[str]:
        """Defaults to ``'Data range: {range} minutes.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._time_range_minutes

    @time_range_minutes.setter
    def time_range_minutes(self, value):
        self._time_range_minutes = validators.string(value, allow_empty = True)

    @property
    def time_range_seconds(self) -> Optional[str]:
        """Defaults to ``'Data range: {range} seconds.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._time_range_seconds

    @time_range_seconds.setter
    def time_range_seconds(self, value):
        self._time_range_seconds = validators.string(value, allow_empty = True)

    @property
    def x_axis_description_plural(self) -> Optional[str]:
        """Defaults to ``'The chart has {numAxes} X axes displaying {#each(names, -1), }and {names[-1]}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._x_axis_description_plural

    @x_axis_description_plural.setter
    def x_axis_description_plural(self, value):
        self._x_axis_description_plural = validators.string(value, allow_empty = True)

    @property
    def x_axis_description_singular(self) -> Optional[str]:
        """Defaults to ``'The chart has 1 X axis displaying {names[0]}. {ranges[0]}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._x_axis_description_singular

    @x_axis_description_singular.setter
    def x_axis_description_singular(self, value):
        self._x_axis_description_singular = validators.string(value, allow_empty = True)

    @property
    def y_axis_description_plural(self) -> Optional[str]:
        """Defaults to ``'The chart has {numAxes} Y axes displaying {#each(names, -1), }and {names[-1]}.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._y_axis_description_plural

    @y_axis_description_plural.setter
    def y_axis_description_plural(self, value):
        self._y_axis_description_plural = validators.string(value, allow_empty = True)

    @property
    def y_axis_description_singular(self) -> Optional[str]:
        """Defaults to ``'The chart has 1 Y axis displaying {names[0]}. {ranges[0]}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._y_axis_description_singular

    @y_axis_description_singular.setter
    def y_axis_description_singular(self, value):
        self._y_axis_description_singular = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'range_categories': as_dict.get('rangeCategories', None),
            'range_from_to': as_dict.get('rangeFromTo', None),
            'time_range_days': as_dict.get('timeRangeDays', None),
            'time_range_hours': as_dict.get('timeRangeHours', None),
            'time_range_minutes': as_dict.get('timeRangeMinutes', None),
            'time_range_seconds': as_dict.get('timeRangeSeconds', None),
            'x_axis_description_plural': as_dict.get('xAxisDescriptionPlural', None),
            'x_axis_description_singular': as_dict.get('xAxisDescriptionSingular', None),
            'y_axis_description_plural': as_dict.get('yAxisDescriptionPlural', None),
            'y_axis_description_singular': as_dict.get('yAxisDescriptionSingular', None),
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'rangeCategories': self.range_categories,
            'rangeFromTo': self.range_from_to,
            'timeRangeDays': self.time_range_days,
            'timeRangeHours': self.time_range_hours,
            'timeRangeMinutes': self.time_range_minutes,
            'timeRangeSeconds': self.time_range_seconds,
            'xAxisDescriptionPlural': self.x_axis_description_plural,
            'xAxisDescriptionSingular': self.x_axis_description_singular,
            'yAxisDescriptionPlural': self.y_axis_description_plural,
            'yAxisDescriptionSingular': self.y_axis_description_singular
        }

        return untrimmed
